from typing import Any, Optional

from gi.types import GObjectMeta


class SingletonGObjectMeta(GObjectMeta):
    _instance: Optional[Any] = None

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if not cls._instance:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance



from gettext import gettext as _
from typing import Optional

from blueman.gui.GtkAnimation import WidgetFade

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Pango


class MessageArea(Gtk.InfoBar):
    _inst_: "MessageArea"

    def __new__(cls) -> "MessageArea":
        if not hasattr(MessageArea, "_inst_"):
            MessageArea._inst_ = super().__new__(cls)

        return MessageArea._inst_

    def __init__(self) -> None:
        super().__init__(show_close_button=True)

        self.set_name("MessageArea")

        self.text = ""

        self.anim = WidgetFade(self, self.get_style_context().get_background_color(Gtk.StateFlags.NORMAL))
        self.hl_anim = WidgetFade(self, Gdk.RGBA(1, 0, 0, 1))

        self.icon = Gtk.Image(pixel_size=16, visible=True)
        self.label = Gtk.Label(xalign=0, ellipsize=Pango.EllipsizeMode.END, single_line_mode=True,
                               selectable=True, visible=True)

        im = Gtk.Image(icon_name="dialog-information", pixel_size=16, visible=True)
        self.b_more = self.add_button(_("More"), 0)
        self.b_more.set_image(im)
        self.b_more.props.relief = Gtk.ReliefStyle.NONE

        self.content_area = self.get_content_area()
        self.content_area.add(self.icon)
        self.content_area.add(self.label)

        self.connect("response", self.on_response)

    def on_response(self, info_bar: Gtk.InfoBar, response_id: int) -> None:
        if response_id == 0:
            assert self.bt is not None
            parent = self.get_toplevel()
            assert isinstance(parent, Gtk.Container)
            d = Gtk.MessageDialog(parent=parent, type=Gtk.MessageType.INFO,
                                  buttons=Gtk.ButtonsType.CLOSE, text='\n'.join((self.text, self.bt)))
            d.run()
            d.destroy()
        elif response_id == Gtk.ResponseType.CLOSE:
            info_bar.props.visible = False

    @staticmethod
    def close() -> None:
        MessageArea._inst_.response(1)

    @staticmethod
    def show_message(text: str, bt: Optional[str] = None, icon: str = "dialog-warning") -> None:
        MessageArea._inst_._show_message(text, bt, icon)

    def _show_message(self, text: str, bt: Optional[str] = None, icon: str = "dialog-warning") -> None:
        def on_finished(anim: WidgetFade) -> None:
            anim.disconnect(sig)
            anim.freeze()

        self.text = text
        self.bt = bt

        self.label.props.tooltip_text = text
        self.icon.props.icon_name = icon

        if icon == "dialog-warning":
            self.set_message_type(Gtk.MessageType.ERROR)
            self.hl_anim.color = Gdk.RGBA(1, 0, 0, 1)
        else:
            self.set_message_type(Gtk.MessageType.INFO)
            self.hl_anim.color = Gdk.RGBA(0, 0, 1, 1)

        if not self.props.visible:
            sig = self.anim.connect("animation-finished", on_finished)
            self.anim.thaw()
            self.show()
            self.anim.animate(start=0.0, end=1.0, duration=500)
        else:
            sig = self.hl_anim.connect("animation-finished", on_finished)
            self.hl_anim.thaw()
            self.hl_anim.animate(start=0.7, end=1.0, duration=1000)

        if self.bt:
            self.label.props.label = self.text + "…"
            self.b_more.props.visible = True
        else:
            self.label.props.label = self.text
            self.b_more.props.visible = False

        # Queue a resize to workaround negative height problem, see issue #369
        # TODO revisit this in later Gtk versions to see if the problem persists
        self.queue_resize()



import logging

from blueman.bluez.errors import BluezDBusException
from blueman.bluez.obex.Base import Base
from gi.repository import GLib


class AgentManager(Base):
    _interface_name = 'org.bluez.obex.AgentManager1'
    _obj_path = '/org/bluez/obex'

    def __init__(self) -> None:
        super().__init__(obj_path=self._obj_path)

    def register_agent(self, agent_path: str) -> None:
        def on_registered() -> None:
            logging.info(agent_path)

        def on_register_failed(error: BluezDBusException) -> None:
            logging.error(f"{agent_path} {error}")

        param = GLib.Variant('(o)', (agent_path,))
        self._call('RegisterAgent', param, reply_handler=on_registered, error_handler=on_register_failed)

    def unregister_agent(self, agent_path: str) -> None:
        def on_unregistered() -> None:
            logging.info(agent_path)

        def on_unregister_failed(error: BluezDBusException) -> None:
            logging.error(f"{agent_path} {error}")

        param = GLib.Variant('(o)', (agent_path,))
        self._call('UnregisterAgent', param, reply_handler=on_unregistered, error_handler=on_unregister_failed)



import errno
import logging
from enum import Enum, auto
from gettext import gettext as _
from typing import Dict, List, Tuple, Optional, TYPE_CHECKING, Union, Iterable, Callable

from blueman.Functions import create_menuitem, e_
from blueman.bluez.Network import AnyNetwork
from blueman.bluez.Device import AnyDevice, Device
from blueman.config.AutoConnectConfig import AutoConnectConfig
from blueman.gui.manager.ManagerProgressbar import ManagerProgressbar
from blueman.main.Builder import Builder
from blueman.main.DBusProxies import AppletService, DBusProxyFailed
from blueman.gui.MessageArea import MessageArea
from blueman.Sdp import (
    ServiceUUID,
    AUDIO_SOURCE_SVCLASS_ID,
    AUDIO_SINK_SVCLASS_ID,
    HANDSFREE_AGW_SVCLASS_ID,
    HANDSFREE_SVCLASS_ID,
    HEADSET_SVCLASS_ID,
    HID_SVCLASS_ID)

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk
from gi.repository import GLib

if TYPE_CHECKING:
    from blueman.gui.manager.ManagerDeviceList import ManagerDeviceList
    from blueman.main.Manager import Blueman


class DeviceMenuItem:
    class Group(Enum):
        CONNECT = auto()
        DISCONNECT = auto()
        AUTOCONNECT = auto()
        ACTIONS = auto()

    def __init__(self, item: Gtk.MenuItem, group: Group, position: int):
        self.item = item
        self.group = group
        self.position = position


class MenuItemsProvider:
    def on_request_menu_items(self, manager_menu: "ManagerDeviceMenu", device: Device) -> List[DeviceMenuItem]:
        ...


class ManagerDeviceMenu(Gtk.Menu):
    __ops__: Dict[str, str] = {}
    __instances__: List["ManagerDeviceMenu"] = []

    SelectedDevice: Device

    def __init__(self, blueman: "Blueman") -> None:
        super().__init__()
        self.set_name("ManagerDeviceMenu")
        self.Blueman = blueman

        self.is_popup = False

        self._device_property_changed_signal = self.Blueman.List.connect("device-property-changed",
                                                                         self.on_device_property_changed)
        ManagerDeviceMenu.__instances__.append(self)

        self._any_network = AnyNetwork()
        self._any_network.connect_signal('property-changed', self._on_service_property_changed)

        self._any_device = AnyDevice()
        self._any_device.connect_signal('property-changed', self._on_service_property_changed)

        try:
            self._appl: Optional[AppletService] = AppletService()
        except DBusProxyFailed:
            logging.error("** Failed to connect to applet", exc_info=True)
            self._appl = None

        self.generate()

    def __del__(self) -> None:
        logging.debug("deleting devicemenu")

    def popup_at_pointer(self, event: Optional[Gdk.Event]) -> None:
        self.is_popup = True
        self.generate()

        super().popup_at_pointer(event)

    def clear(self) -> None:
        def remove_and_destroy(child: Gtk.Widget) -> None:
            self.remove(child)
            child.destroy()

        self.foreach(remove_and_destroy)

    def set_op(self, device: Device, message: str) -> None:
        ManagerDeviceMenu.__ops__[device.get_object_path()] = message
        for inst in ManagerDeviceMenu.__instances__:
            logging.info(f"op: regenerating instance {inst}")
            if inst.SelectedDevice == self.SelectedDevice and not (inst.is_popup and not inst.props.visible):
                inst.generate()

    def get_op(self, device: Device) -> Optional[str]:
        try:
            return ManagerDeviceMenu.__ops__[device.get_object_path()]
        except KeyError:
            return None

    def unset_op(self, device: Device) -> None:
        del ManagerDeviceMenu.__ops__[device.get_object_path()]
        for inst in ManagerDeviceMenu.__instances__:
            logging.info(f"op: regenerating instance {inst}")
            if inst.SelectedDevice == self.SelectedDevice and not (inst.is_popup and not inst.props.visible):
                inst.generate()

    def _on_service_property_changed(self, _service: Union[AnyNetwork, AnyDevice], key: str, _value: object,
                                     _path: str) -> None:
        if key == "Connected":
            self.generate()

    GENERIC_CONNECT = "00000000-0000-0000-0000-000000000000"

    def connect_service(self, device: Device, uuid: str = GENERIC_CONNECT) -> None:
        def success(_obj: AppletService, _result: None, _user_data: None) -> None:
            logging.info("success")
            prog.message(_("Success!"))

            MessageArea.close()

            self.unset_op(device)

        def fail(_obj: Optional[AppletService], result: GLib.Error, _user_data: None) -> None:
            prog.message(_("Failed"))

            self.unset_op(device)
            logging.warning(f"fail {result}")
            self._handle_error_message(result)

        self.set_op(device, _("Connecting…"))
        prog = ManagerProgressbar(self.Blueman, cancellable=uuid == self.GENERIC_CONNECT)
        if uuid == self.GENERIC_CONNECT:
            prog.connect("cancelled", lambda x: self.disconnect_service(device))

        if self._appl is None:
            fail(None, GLib.Error('Applet DBus Service not available'), None)
            return

        def connect(error_handler: Callable[[AppletService, GLib.Error, None], None]) -> None:
            assert self._appl is not None  # https://github.com/python/mypy/issues/2608
            self._appl.ConnectService('(os)', device.get_object_path(), uuid,
                                      result_handler=success, error_handler=error_handler,
                                      timeout=GLib.MAXINT)

        def initial_error_handler(obj: AppletService, result: GLib.Error, user_date: None) -> None:
            # There are (Intel) drivers that fail to connect while a discovery is running
            if self._get_errno(result) == errno.EAGAIN:
                assert self.Blueman.List.Adapter is not None
                self.Blueman.List.Adapter.stop_discovery()
                connect(fail)
            else:
                fail(obj, result, user_date)

        connect(initial_error_handler)

        prog.start()

    def disconnect_service(self, device: Device, uuid: str = GENERIC_CONNECT, port: int = 0) -> None:
        def ok(_obj: AppletService, _result: None, _user_date: None) -> None:
            logging.info("disconnect success")
            self.generate()

        def err(_obj: Optional[AppletService], result: GLib.Error, _user_date: None) -> None:
            logging.warning(f"disconnect failed {result}")
            msg, tb = e_(result.message)
            MessageArea.show_message(_("Disconnection Failed: ") + msg, tb)
            self.generate()

        if self._appl is None:
            err(None, GLib.Error('Applet DBus Service not available'), None)
            return

        self._appl.DisconnectService('(osd)', device.get_object_path(), uuid, port,
                                     result_handler=ok, error_handler=err, timeout=GLib.MAXINT)

    def on_device_property_changed(self, lst: "ManagerDeviceList", _device: Device, tree_iter: Gtk.TreeIter,
                                   key_value: Tuple[str, object]) -> None:
        key, value = key_value
        # print "menu:", key, value
        if lst.compare(tree_iter, lst.selected()):
            if key in ("Connected", "UUIDs", "Trusted", "Paired"):
                self.generate()

    def _handle_error_message(self, error: GLib.Error) -> None:
        err = self._get_errno(error)

        if err == errno.ENOPROTOOPT:
            logging.warning("No audio endpoints registered to bluetoothd. "
                            "Pulseaudio Bluetooth module, bluez-alsa, PipeWire or other audio support missing.")
            msg = _("No audio endpoints registered")
        elif err == errno.EIO:
            logging.warning("bluetoothd reported input/output error. Check its logs for context.")
            msg = _("Input/output error")
        elif err == errno.EHOSTDOWN:
            msg = _("Device did not respond")
        elif err == errno.EAGAIN:
            logging.warning("bluetoothd reported resource temporarily unavailable. "
                            "Retry or check its logs for context.")
            msg = _("Resource temporarily unavailable")
        else:
            msg = error.message.split(":", 3)[-1].strip()

        if msg != "Cancelled":
            MessageArea.show_message(_("Connection Failed: ") + msg)

    @staticmethod
    def _get_errno(error: GLib.Error) -> Optional[int]:
        msg = error.message.split(":", 3)[-1].strip()

        # https://sourceware.org/git/?p=glibc.git;a=blob;f=sysdeps/gnu/errlist.h
        # https://git.musl-libc.org/cgit/musl/tree/src/errno/__strerror.h
        # https://git.uclibc.org/uClibc/tree/libc/string/_string_syserrmsgs.c
        if msg == "Protocol not available":
            return errno.ENOPROTOOPT
        if msg in ("Input/output error", "I/O error"):
            return errno.EIO
        if msg == "Host is down":
            # Bluetooth errors 0x04 (Page Timeout) or 0x3c (Advertising Timeout)
            return errno.EHOSTDOWN
        if msg == "Resource temporarily unavailable":
            return errno.EAGAIN

        return None

    def show_generic_connect_calc(self, device_uuids: Iterable[str]) -> bool:
        # Generic (dis)connect
        for uuid in device_uuids:
            service_uuid = ServiceUUID(uuid)
            if service_uuid.short_uuid in (
                    AUDIO_SOURCE_SVCLASS_ID, AUDIO_SINK_SVCLASS_ID, HANDSFREE_AGW_SVCLASS_ID, HANDSFREE_SVCLASS_ID,
                    HEADSET_SVCLASS_ID, HID_SVCLASS_ID, 0x1812
            ):
                return True
            elif not service_uuid.reserved:
                if uuid == '03b80e5a-ede8-4b33-a751-6ce34ec4c700':
                    return True
        # LE devices do not appear to expose certain properties like uuids until connect to at least once.
        return not device_uuids

    def generate(self) -> None:
        self.clear()

        if not self.is_popup or self.props.visible:
            selected = self.Blueman.List.selected()
            if not selected:
                return
            row = self.Blueman.List.get(selected, "alias", "paired", "connected", "trusted", "objpush", "device")
        else:
            (x, y) = self.Blueman.List.get_pointer()
            path = self.Blueman.List.get_path_at_pos(x, y)
            if path is not None:
                assert path[0] is not None
                row = self.Blueman.List.get(path[0], "alias", "paired", "connected", "trusted", "objpush", "device")
            else:
                return

        self.SelectedDevice = row["device"]

        op = self.get_op(self.SelectedDevice)

        if op is not None:
            item: Gtk.MenuItem = create_menuitem(op, "network-transmit-receive")
            item.props.sensitive = False
            item.show()
            self.append(item)
            return

        show_generic_connect = self.show_generic_connect_calc(self.SelectedDevice['UUIDs'])

        if not row["connected"] and show_generic_connect:
            connect_item = create_menuitem(_("_<b>_Connect</b>"), "blueman")
            connect_item.connect("activate", lambda _item: self.connect_service(self.SelectedDevice))
            connect_item.props.tooltip_text = _("Connects auto connect profiles A2DP source, A2DP sink, and HID")
            connect_item.show()
            self.append(connect_item)
        elif show_generic_connect:
            connect_item = create_menuitem(_("_<b>Disconnect</b>"), "network-offline")
            connect_item.props.tooltip_text = _("Forcefully disconnect the device")
            connect_item.connect("activate", lambda _item: self.disconnect_service(self.SelectedDevice))
            connect_item.show()
            self.append(connect_item)

        logging.debug(row["alias"])

        items = [item for plugin in self.Blueman.Plugins.get_loaded_plugins(MenuItemsProvider)
                 for item in plugin.on_request_menu_items(self, self.SelectedDevice)]

        connect_items = [i for i in items if i.group == DeviceMenuItem.Group.CONNECT]
        disconnect_items = [i for i in items if i.group == DeviceMenuItem.Group.DISCONNECT]
        autoconnect_items = [item for item in items if item.group == DeviceMenuItem.Group.AUTOCONNECT]
        action_items = [i for i in items if i.group == DeviceMenuItem.Group.ACTIONS]

        if connect_items:
            self.append(self._create_header(_("<b>Connect To:</b>")))
            for it in sorted(connect_items, key=lambda i: i.position):
                self.append(it.item)

        if disconnect_items:
            self.append(self._create_header(_("<b>Disconnect:</b>")))
            for it in sorted(disconnect_items, key=lambda i: i.position):
                self.append(it.item)

        config = AutoConnectConfig()
        generic_service = ServiceUUID("00000000-0000-0000-0000-000000000000")
        generic_autoconnect = (self.SelectedDevice.get_object_path(), str(generic_service)) in set(config["services"])

        if row["connected"] or generic_autoconnect or autoconnect_items:
            self.append(self._create_header(_("<b>Auto-connect:</b>")))

            if row["connected"] or generic_autoconnect:
                item = Gtk.CheckMenuItem(label=generic_service.name)
                config.bind_to_menuitem(item, self.SelectedDevice, str(generic_service))
                item.show()
                self.append(item)

            for it in sorted(autoconnect_items, key=lambda i: i.position):
                self.append(it.item)

        if show_generic_connect or connect_items or disconnect_items or autoconnect_items:
            item = Gtk.SeparatorMenuItem()
            item.show()
            self.append(item)

        for it in sorted(action_items, key=lambda i: i.position):
            self.append(it.item)

        send_item = create_menuitem(_("Send a _File…"), "edit-copy")
        send_item.props.sensitive = False
        self.append(send_item)
        send_item.show()

        if row["objpush"]:
            send_item.connect("activate", lambda x: self.Blueman.send(self.SelectedDevice))
            send_item.props.sensitive = True

        item = Gtk.SeparatorMenuItem()
        item.show()
        self.append(item)

        item = create_menuitem(_("_Pair"), "dialog-password")
        item.props.tooltip_text = _("Create pairing with the device")
        self.append(item)
        item.show()
        if not row["paired"]:
            item.connect("activate", lambda x: self.Blueman.bond(self.SelectedDevice))
        else:
            item.props.sensitive = False

        if not row["trusted"]:
            item = create_menuitem(_("_Trust"), "blueman-trust")
            item.connect("activate", lambda x: self.Blueman.toggle_trust(self.SelectedDevice))
            self.append(item)
            item.show()
        else:
            item = create_menuitem(_("_Untrust"), "blueman-untrust")
            self.append(item)
            item.connect("activate", lambda x: self.Blueman.toggle_trust(self.SelectedDevice))
            item.show()
        item.props.tooltip_text = _("Mark/Unmark this device as trusted")

        def on_rename(_item: Gtk.MenuItem, device: Device) -> None:
            def on_response(dialog: Gtk.Dialog, response_id: int) -> None:
                if response_id == Gtk.ResponseType.ACCEPT:
                    assert isinstance(alias_entry, Gtk.Entry)  # https://github.com/python/mypy/issues/2608
                    device.set('Alias', alias_entry.get_text())
                elif response_id == 1:
                    device.set('Alias', '')
                dialog.destroy()

            builder = Builder("rename-device.ui")
            dialog = builder.get_widget("dialog", Gtk.Dialog)
            dialog.set_transient_for(self.Blueman.window)
            dialog.props.icon_name = "blueman"
            alias_entry = builder.get_widget("alias_entry", Gtk.Entry)
            alias_entry.set_text(device['Alias'])
            dialog.connect("response", on_response)
            dialog.present()

        item = Gtk.MenuItem.new_with_mnemonic(_("R_ename device…"))
        item.connect('activate', on_rename, self.SelectedDevice)
        self.append(item)
        item.show()

        item = Gtk.SeparatorMenuItem()
        item.show()
        self.append(item)

        item = create_menuitem(_("_Remove…"), "edit-delete")
        item.connect("activate", lambda x: self.Blueman.remove(self.SelectedDevice))
        self.append(item)
        item.show()
        item.props.tooltip_text = _("Remove this device from the known devices list")

    @staticmethod
    def _create_header(text: str) -> Gtk.MenuItem:
        item = Gtk.MenuItem()
        label = Gtk.Label()
        label.set_markup(text)
        label.props.xalign = 0.0

        label.show()
        item.add(label)
        item.props.sensitive = False
        item.show()
        return item



from gettext import gettext as _
from typing import Callable, Union

from gi.repository import GLib

from blueman.Service import Service
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.NetworkManager import NMDUNConnection, NMConnectionError
from blueman.Sdp import DIALUP_NET_SVCLASS_ID
from blueman.plugins.applet.DBusService import ServiceConnectHandler


class NMDUNSupport(AppletPlugin, ServiceConnectHandler):
    __depends__ = ["StatusIcon", "DBusService"]
    __conflicts__ = ["PPPSupport"]
    __icon__ = "modem"
    __author__ = "infirit"
    __description__ = _("Provides support for Dial Up Networking (DUN) with ModemManager and NetworkManager")
    __priority__ = 1

    def service_connect_handler(self, service: Service, ok: Callable[[], None],
                                err: Callable[[Union[NMConnectionError, GLib.Error]], None]) -> bool:
        if DIALUP_NET_SVCLASS_ID != service.short_uuid:
            return False

        conn = NMDUNConnection(service, ok, err)
        conn.activate()

        return True

    def service_disconnect_handler(self, service: Service, ok: Callable[[], None],
                                   err: Callable[[Union[NMConnectionError, GLib.Error]], None]) -> bool:
        if DIALUP_NET_SVCLASS_ID != service.short_uuid:
            return False

        conn = NMDUNConnection(service, ok, err)
        conn.deactivate()

        return True



from typing import Any, Optional

from gi.repository import Gio, GObject
import logging


class Config(Gio.Settings):
    def __init__(self, schema_id: str, path: Optional[str] = None) -> None:
        # Add backwards compat with pygobject < 3.11.2
        if Gio.Settings.__init__.__name__ == "new_init":
            super().__init__(schema_id=schema_id, path=path)
        else:
            super().__init__(schema=schema_id, path=path)

    def bind_to_widget(self, key: str, widget: GObject.Object, prop: str,
                       flags: Gio.SettingsBindFlags = Gio.SettingsBindFlags.DEFAULT) -> None:
        self.bind(key, widget, prop, flags)

    def __setitem__(self, key: str, value: Any) -> None:
        if not self.is_writable(key):
            logging.error("GSetting not writable, settings not saved")
        super().__setitem__(key, value)



import builtins
import typing

from gi.repository import GLib
from gi.repository import GObject


class Action(GObject.GInterface):

    def do_action(self, i: builtins.int) -> builtins.bool: ...

    def get_description(self, i: builtins.int) -> typing.Optional[builtins.str]: ...

    def get_keybinding(self, i: builtins.int) -> typing.Optional[builtins.str]: ...

    def get_localized_name(self, i: builtins.int) -> typing.Optional[builtins.str]: ...

    def get_n_actions(self) -> builtins.int: ...

    def get_name(self, i: builtins.int) -> typing.Optional[builtins.str]: ...

    def set_description(self, i: builtins.int, desc: builtins.str) -> builtins.bool: ...

    def do_do_action(self, i: builtins.int) -> builtins.bool: ...

    def do_get_description(self, i: builtins.int) -> typing.Optional[builtins.str]: ...

    def do_get_keybinding(self, i: builtins.int) -> typing.Optional[builtins.str]: ...

    def do_get_localized_name(self, i: builtins.int) -> typing.Optional[builtins.str]: ...

    def do_get_n_actions(self) -> builtins.int: ...

    def do_get_name(self, i: builtins.int) -> typing.Optional[builtins.str]: ...

    def do_set_description(self, i: builtins.int, desc: builtins.str) -> builtins.bool: ...


class Component(GObject.GInterface):

    def contains(self, x: builtins.int, y: builtins.int, coord_type: CoordType) -> builtins.bool: ...

    def get_alpha(self) -> builtins.float: ...

    def get_extents(self, coord_type: CoordType) -> typing.Tuple[builtins.int, builtins.int, builtins.int, builtins.int]: ...

    def get_layer(self) -> Layer: ...

    def get_mdi_zorder(self) -> builtins.int: ...

    def get_position(self, coord_type: CoordType) -> typing.Tuple[builtins.int, builtins.int]: ...

    def get_size(self) -> typing.Tuple[builtins.int, builtins.int]: ...

    def grab_focus(self) -> builtins.bool: ...

    def ref_accessible_at_point(self, x: builtins.int, y: builtins.int, coord_type: CoordType) -> typing.Optional[Object]: ...

    def remove_focus_handler(self, handler_id: builtins.int) -> None: ...

    def scroll_to(self, type: ScrollType) -> builtins.bool: ...

    def scroll_to_point(self, coords: CoordType, x: builtins.int, y: builtins.int) -> builtins.bool: ...

    def set_extents(self, x: builtins.int, y: builtins.int, width: builtins.int, height: builtins.int, coord_type: CoordType) -> builtins.bool: ...

    def set_position(self, x: builtins.int, y: builtins.int, coord_type: CoordType) -> builtins.bool: ...

    def set_size(self, width: builtins.int, height: builtins.int) -> builtins.bool: ...

    def do_bounds_changed(self, bounds: Rectangle) -> None: ...

    def do_contains(self, x: builtins.int, y: builtins.int, coord_type: CoordType) -> builtins.bool: ...

    def do_get_alpha(self) -> builtins.float: ...

    def do_get_extents(self, coord_type: CoordType) -> typing.Tuple[builtins.int, builtins.int, builtins.int, builtins.int]: ...

    def do_get_layer(self) -> Layer: ...

    def do_get_mdi_zorder(self) -> builtins.int: ...

    def do_get_position(self, coord_type: CoordType) -> typing.Tuple[builtins.int, builtins.int]: ...

    def do_get_size(self) -> typing.Tuple[builtins.int, builtins.int]: ...

    def do_grab_focus(self) -> builtins.bool: ...

    def do_ref_accessible_at_point(self, x: builtins.int, y: builtins.int, coord_type: CoordType) -> typing.Optional[Object]: ...

    def do_remove_focus_handler(self, handler_id: builtins.int) -> None: ...

    def do_scroll_to(self, type: ScrollType) -> builtins.bool: ...

    def do_scroll_to_point(self, coords: CoordType, x: builtins.int, y: builtins.int) -> builtins.bool: ...

    def do_set_extents(self, x: builtins.int, y: builtins.int, width: builtins.int, height: builtins.int, coord_type: CoordType) -> builtins.bool: ...

    def do_set_position(self, x: builtins.int, y: builtins.int, coord_type: CoordType) -> builtins.bool: ...

    def do_set_size(self, width: builtins.int, height: builtins.int) -> builtins.bool: ...


class Document(GObject.GInterface):

    def get_attribute_value(self, attribute_name: builtins.str) -> typing.Optional[builtins.str]: ...

    def get_attributes(self) -> typing.Sequence[builtins.object]: ...

    def get_current_page_number(self) -> builtins.int: ...

    def get_document(self) -> typing.Optional[builtins.object]: ...

    def get_document_type(self) -> builtins.str: ...

    def get_locale(self) -> builtins.str: ...

    def get_page_count(self) -> builtins.int: ...

    def set_attribute_value(self, attribute_name: builtins.str, attribute_value: builtins.str) -> builtins.bool: ...

    def do_get_current_page_number(self) -> builtins.int: ...

    def do_get_document(self) -> typing.Optional[builtins.object]: ...

    def do_get_document_attribute_value(self, attribute_name: builtins.str) -> typing.Optional[builtins.str]: ...

    def do_get_document_attributes(self) -> typing.Sequence[builtins.object]: ...

    def do_get_document_locale(self) -> builtins.str: ...

    def do_get_document_type(self) -> builtins.str: ...

    def do_get_page_count(self) -> builtins.int: ...

    def do_set_document_attribute(self, attribute_name: builtins.str, attribute_value: builtins.str) -> builtins.bool: ...


class EditableText(GObject.GInterface):

    def copy_text(self, start_pos: builtins.int, end_pos: builtins.int) -> None: ...

    def cut_text(self, start_pos: builtins.int, end_pos: builtins.int) -> None: ...

    def delete_text(self, start_pos: builtins.int, end_pos: builtins.int) -> None: ...

    def insert_text(self, string: builtins.str, length: builtins.int, position: builtins.int) -> None: ...

    def paste_text(self, position: builtins.int) -> None: ...

    def set_run_attributes(self, attrib_set: typing.Sequence[builtins.object], start_offset: builtins.int, end_offset: builtins.int) -> builtins.bool: ...

    def set_text_contents(self, string: builtins.str) -> None: ...

    def do_copy_text(self, start_pos: builtins.int, end_pos: builtins.int) -> None: ...

    def do_cut_text(self, start_pos: builtins.int, end_pos: builtins.int) -> None: ...

    def do_delete_text(self, start_pos: builtins.int, end_pos: builtins.int) -> None: ...

    def do_insert_text(self, string: builtins.str, length: builtins.int, position: builtins.int) -> None: ...

    def do_paste_text(self, position: builtins.int) -> None: ...

    def do_set_run_attributes(self, attrib_set: typing.Sequence[builtins.object], start_offset: builtins.int, end_offset: builtins.int) -> builtins.bool: ...

    def do_set_text_contents(self, string: builtins.str) -> None: ...


class HyperlinkImpl(GObject.GInterface):

    def get_hyperlink(self) -> Hyperlink: ...

    def do_get_hyperlink(self) -> Hyperlink: ...


class Hypertext(GObject.GInterface):

    def get_link(self, link_index: builtins.int) -> Hyperlink: ...

    def get_link_index(self, char_index: builtins.int) -> builtins.int: ...

    def get_n_links(self) -> builtins.int: ...

    def do_get_link(self, link_index: builtins.int) -> Hyperlink: ...

    def do_get_link_index(self, char_index: builtins.int) -> builtins.int: ...

    def do_get_n_links(self) -> builtins.int: ...

    def do_link_selected(self, link_index: builtins.int) -> None: ...


class Image(GObject.GInterface):

    def get_image_description(self) -> builtins.str: ...

    def get_image_locale(self) -> typing.Optional[builtins.str]: ...

    def get_image_position(self, coord_type: CoordType) -> typing.Tuple[builtins.int, builtins.int]: ...

    def get_image_size(self) -> typing.Tuple[builtins.int, builtins.int]: ...

    def set_image_description(self, description: builtins.str) -> builtins.bool: ...

    def do_get_image_description(self) -> builtins.str: ...

    def do_get_image_locale(self) -> typing.Optional[builtins.str]: ...

    def do_get_image_position(self, coord_type: CoordType) -> typing.Tuple[builtins.int, builtins.int]: ...

    def do_get_image_size(self) -> typing.Tuple[builtins.int, builtins.int]: ...

    def do_set_image_description(self, description: builtins.str) -> builtins.bool: ...


class ImplementorIface(GObject.GInterface):
    ...


class Misc(GObject.Object):
    parent: GObject.Object

    @staticmethod
    def get_instance() -> Misc: ...

    def threads_enter(self) -> None: ...

    def threads_leave(self) -> None: ...

    def do_threads_enter(self) -> None: ...

    def do_threads_leave(self) -> None: ...


class Object(GObject.Object):
    accessible_parent: Object
    description: builtins.str
    layer: Layer
    name: builtins.str
    parent: GObject.Object
    relation_set: RelationSet
    role: Role

    def add_relationship(self, relationship: RelationType, target: Object) -> builtins.bool: ...

    def get_accessible_id(self) -> builtins.str: ...

    def get_attributes(self) -> typing.Sequence[builtins.object]: ...

    def get_description(self) -> builtins.str: ...

    def get_index_in_parent(self) -> builtins.int: ...

    def get_layer(self) -> Layer: ...

    def get_mdi_zorder(self) -> builtins.int: ...

    def get_n_accessible_children(self) -> builtins.int: ...

    def get_name(self) -> builtins.str: ...

    def get_object_locale(self) -> builtins.str: ...

    def get_parent(self) -> Object: ...

    def get_role(self) -> Role: ...

    def initialize(self, data: typing.Optional[builtins.object]) -> None: ...

    def notify_state_change(self, state: builtins.int, value: builtins.bool) -> None: ...

    def peek_parent(self) -> Object: ...

    def ref_accessible_child(self, i: builtins.int) -> Object: ...

    def ref_relation_set(self) -> RelationSet: ...

    def ref_state_set(self) -> StateSet: ...

    def remove_property_change_handler(self, handler_id: builtins.int) -> None: ...

    def remove_relationship(self, relationship: RelationType, target: Object) -> builtins.bool: ...

    def set_accessible_id(self, name: builtins.str) -> None: ...

    def set_description(self, description: builtins.str) -> None: ...

    def set_name(self, name: builtins.str) -> None: ...

    def set_parent(self, parent: Object) -> None: ...

    def set_role(self, role: Role) -> None: ...

    def do_active_descendant_changed(self, child: typing.Optional[builtins.object]) -> None: ...

    def do_children_changed(self, change_index: builtins.int, changed_child: typing.Optional[builtins.object]) -> None: ...

    def do_focus_event(self, focus_in: builtins.bool) -> None: ...

    def do_get_attributes(self) -> typing.Sequence[builtins.object]: ...

    def do_get_description(self) -> builtins.str: ...

    def do_get_index_in_parent(self) -> builtins.int: ...

    def do_get_layer(self) -> Layer: ...

    def do_get_mdi_zorder(self) -> builtins.int: ...

    def do_get_n_children(self) -> builtins.int: ...

    def do_get_name(self) -> builtins.str: ...

    def do_get_object_locale(self) -> builtins.str: ...

    def do_get_parent(self) -> Object: ...

    def do_get_role(self) -> Role: ...

    def do_initialize(self, data: typing.Optional[builtins.object]) -> None: ...

    def do_property_change(self, values: PropertyValues) -> None: ...

    def do_ref_relation_set(self) -> RelationSet: ...

    def do_ref_state_set(self) -> StateSet: ...

    def do_remove_property_change_handler(self, handler_id: builtins.int) -> None: ...

    def do_set_description(self, description: builtins.str) -> None: ...

    def do_set_name(self, name: builtins.str) -> None: ...

    def do_set_parent(self, parent: Object) -> None: ...

    def do_set_role(self, role: Role) -> None: ...

    def do_state_change(self, name: builtins.str, state_set: builtins.bool) -> None: ...

    def do_visible_data_changed(self) -> None: ...


class ObjectFactory(GObject.Object):
    parent: GObject.Object

    def create_accessible(self, obj: GObject.Object) -> Object: ...

    def get_accessible_type(self) -> GObject.GType: ...

    def invalidate(self) -> None: ...

    def do_invalidate(self) -> None: ...


class Registry(GObject.Object):
    factory_singleton_cache: typing.Mapping[builtins.object, builtins.object]
    factory_type_registry: typing.Mapping[builtins.object, builtins.object]
    parent: GObject.Object

    def get_factory(self, type: GObject.GType) -> ObjectFactory: ...

    def get_factory_type(self, type: GObject.GType) -> GObject.GType: ...

    def set_factory_type(self, type: GObject.GType, factory_type: GObject.GType) -> None: ...


class Relation(GObject.Object):
    parent: GObject.Object
    relationship: RelationType
    target: typing.Sequence[builtins.object]

    def add_target(self, target: Object) -> None: ...

    def get_relation_type(self) -> RelationType: ...

    def get_target(self) -> typing.Sequence[Object]: ...

    @staticmethod
    def new(targets: typing.Sequence[Object], relationship: RelationType, **kwargs) -> Relation: ...  # type: ignore

    def remove_target(self, target: Object) -> builtins.bool: ...


class RelationSet(GObject.Object):
    parent: GObject.Object
    relations: typing.Sequence[builtins.object]

    def add(self, relation: Relation) -> None: ...

    def add_relation_by_type(self, relationship: RelationType, target: Object) -> None: ...

    def contains(self, relationship: RelationType) -> builtins.bool: ...

    def contains_target(self, relationship: RelationType, target: Object) -> builtins.bool: ...

    def get_n_relations(self) -> builtins.int: ...

    def get_relation(self, i: builtins.int) -> Relation: ...

    def get_relation_by_type(self, relationship: RelationType) -> Relation: ...

    @staticmethod
    def new(**kwargs) -> RelationSet: ...  # type: ignore

    def remove(self, relation: Relation) -> None: ...


class Selection(GObject.GInterface):

    def add_selection(self, i: builtins.int) -> builtins.bool: ...

    def clear_selection(self) -> builtins.bool: ...

    def get_selection_count(self) -> builtins.int: ...

    def is_child_selected(self, i: builtins.int) -> builtins.bool: ...

    def ref_selection(self, i: builtins.int) -> typing.Optional[Object]: ...

    def remove_selection(self, i: builtins.int) -> builtins.bool: ...

    def select_all_selection(self) -> builtins.bool: ...

    def do_add_selection(self, i: builtins.int) -> builtins.bool: ...

    def do_clear_selection(self) -> builtins.bool: ...

    def do_get_selection_count(self) -> builtins.int: ...

    def do_is_child_selected(self, i: builtins.int) -> builtins.bool: ...

    def do_ref_selection(self, i: builtins.int) -> typing.Optional[Object]: ...

    def do_remove_selection(self, i: builtins.int) -> builtins.bool: ...

    def do_select_all_selection(self) -> builtins.bool: ...

    def do_selection_changed(self) -> None: ...


class StateSet(GObject.Object):
    parent: GObject.Object

    def add_state(self, type: StateType) -> builtins.bool: ...

    def add_states(self, types: typing.Sequence[StateType]) -> None: ...

    def and_sets(self, compare_set: StateSet) -> StateSet: ...

    def clear_states(self) -> None: ...

    def contains_state(self, type: StateType) -> builtins.bool: ...

    def contains_states(self, types: typing.Sequence[StateType]) -> builtins.bool: ...

    def is_empty(self) -> builtins.bool: ...

    @staticmethod
    def new(**kwargs) -> StateSet: ...  # type: ignore

    def or_sets(self, compare_set: StateSet) -> typing.Optional[StateSet]: ...

    def remove_state(self, type: StateType) -> builtins.bool: ...

    def xor_sets(self, compare_set: StateSet) -> StateSet: ...


class StreamableContent(GObject.GInterface):

    def get_mime_type(self, i: builtins.int) -> builtins.str: ...

    def get_n_mime_types(self) -> builtins.int: ...

    def get_stream(self, mime_type: builtins.str) -> GLib.IOChannel: ...

    def get_uri(self, mime_type: builtins.str) -> typing.Optional[builtins.str]: ...

    def do_get_mime_type(self, i: builtins.int) -> builtins.str: ...

    def do_get_n_mime_types(self) -> builtins.int: ...

    def do_get_stream(self, mime_type: builtins.str) -> GLib.IOChannel: ...

    def do_get_uri(self, mime_type: builtins.str) -> typing.Optional[builtins.str]: ...


class Table(GObject.GInterface):

    def add_column_selection(self, column: builtins.int) -> builtins.bool: ...

    def add_row_selection(self, row: builtins.int) -> builtins.bool: ...

    def get_caption(self) -> typing.Optional[Object]: ...

    def get_column_at_index(self, index_: builtins.int) -> builtins.int: ...

    def get_column_description(self, column: builtins.int) -> builtins.str: ...

    def get_column_extent_at(self, row: builtins.int, column: builtins.int) -> builtins.int: ...

    def get_column_header(self, column: builtins.int) -> typing.Optional[Object]: ...

    def get_index_at(self, row: builtins.int, column: builtins.int) -> builtins.int: ...

    def get_n_columns(self) -> builtins.int: ...

    def get_n_rows(self) -> builtins.int: ...

    def get_row_at_index(self, index_: builtins.int) -> builtins.int: ...

    def get_row_description(self, row: builtins.int) -> typing.Optional[builtins.str]: ...

    def get_row_extent_at(self, row: builtins.int, column: builtins.int) -> builtins.int: ...

    def get_row_header(self, row: builtins.int) -> typing.Optional[Object]: ...

    def get_selected_columns(self, selected: builtins.int) -> builtins.int: ...

    def get_selected_rows(self, selected: builtins.int) -> builtins.int: ...

    def get_summary(self) -> Object: ...

    def is_column_selected(self, column: builtins.int) -> builtins.bool: ...

    def is_row_selected(self, row: builtins.int) -> builtins.bool: ...

    def is_selected(self, row: builtins.int, column: builtins.int) -> builtins.bool: ...

    def ref_at(self, row: builtins.int, column: builtins.int) -> Object: ...

    def remove_column_selection(self, column: builtins.int) -> builtins.bool: ...

    def remove_row_selection(self, row: builtins.int) -> builtins.bool: ...

    def set_caption(self, caption: Object) -> None: ...

    def set_column_description(self, column: builtins.int, description: builtins.str) -> None: ...

    def set_column_header(self, column: builtins.int, header: Object) -> None: ...

    def set_row_description(self, row: builtins.int, description: builtins.str) -> None: ...

    def set_row_header(self, row: builtins.int, header: Object) -> None: ...

    def set_summary(self, accessible: Object) -> None: ...

    def do_add_column_selection(self, column: builtins.int) -> builtins.bool: ...

    def do_add_row_selection(self, row: builtins.int) -> builtins.bool: ...

    def do_column_deleted(self, column: builtins.int, num_deleted: builtins.int) -> None: ...

    def do_column_inserted(self, column: builtins.int, num_inserted: builtins.int) -> None: ...

    def do_column_reordered(self) -> None: ...

    def do_get_caption(self) -> typing.Optional[Object]: ...

    def do_get_column_at_index(self, index_: builtins.int) -> builtins.int: ...

    def do_get_column_description(self, column: builtins.int) -> builtins.str: ...

    def do_get_column_extent_at(self, row: builtins.int, column: builtins.int) -> builtins.int: ...

    def do_get_column_header(self, column: builtins.int) -> typing.Optional[Object]: ...

    def do_get_index_at(self, row: builtins.int, column: builtins.int) -> builtins.int: ...

    def do_get_n_columns(self) -> builtins.int: ...

    def do_get_n_rows(self) -> builtins.int: ...

    def do_get_row_at_index(self, index_: builtins.int) -> builtins.int: ...

    def do_get_row_description(self, row: builtins.int) -> typing.Optional[builtins.str]: ...

    def do_get_row_extent_at(self, row: builtins.int, column: builtins.int) -> builtins.int: ...

    def do_get_row_header(self, row: builtins.int) -> typing.Optional[Object]: ...

    def do_get_selected_columns(self, selected: builtins.int) -> builtins.int: ...

    def do_get_selected_rows(self, selected: builtins.int) -> builtins.int: ...

    def do_get_summary(self) -> Object: ...

    def do_is_column_selected(self, column: builtins.int) -> builtins.bool: ...

    def do_is_row_selected(self, row: builtins.int) -> builtins.bool: ...

    def do_is_selected(self, row: builtins.int, column: builtins.int) -> builtins.bool: ...

    def do_model_changed(self) -> None: ...

    def do_ref_at(self, row: builtins.int, column: builtins.int) -> Object: ...

    def do_remove_column_selection(self, column: builtins.int) -> builtins.bool: ...

    def do_remove_row_selection(self, row: builtins.int) -> builtins.bool: ...

    def do_row_deleted(self, row: builtins.int, num_deleted: builtins.int) -> None: ...

    def do_row_inserted(self, row: builtins.int, num_inserted: builtins.int) -> None: ...

    def do_row_reordered(self) -> None: ...

    def do_set_caption(self, caption: Object) -> None: ...

    def do_set_column_description(self, column: builtins.int, description: builtins.str) -> None: ...

    def do_set_column_header(self, column: builtins.int, header: Object) -> None: ...

    def do_set_row_description(self, row: builtins.int, description: builtins.str) -> None: ...

    def do_set_row_header(self, row: builtins.int, header: Object) -> None: ...

    def do_set_summary(self, accessible: Object) -> None: ...


class TableCell(GObject.GInterface):

    def get_column_header_cells(self) -> typing.Sequence[Object]: ...

    def get_column_span(self) -> builtins.int: ...

    def get_position(self) -> typing.Tuple[builtins.bool, builtins.int, builtins.int]: ...

    def get_row_column_span(self) -> typing.Tuple[builtins.bool, builtins.int, builtins.int, builtins.int, builtins.int]: ...

    def get_row_header_cells(self) -> typing.Sequence[Object]: ...

    def get_row_span(self) -> builtins.int: ...

    def get_table(self) -> Object: ...

    def do_get_column_header_cells(self) -> typing.Sequence[Object]: ...

    def do_get_column_span(self) -> builtins.int: ...

    def do_get_position(self) -> typing.Tuple[builtins.bool, builtins.int, builtins.int]: ...

    def do_get_row_column_span(self) -> typing.Tuple[builtins.bool, builtins.int, builtins.int, builtins.int, builtins.int]: ...

    def do_get_row_header_cells(self) -> typing.Sequence[Object]: ...

    def do_get_row_span(self) -> builtins.int: ...

    def do_get_table(self) -> Object: ...


class Text(GObject.GInterface):

    def add_selection(self, start_offset: builtins.int, end_offset: builtins.int) -> builtins.bool: ...

    @staticmethod
    def free_ranges(ranges: typing.Sequence[TextRange]) -> None: ...

    def get_bounded_ranges(self, rect: TextRectangle, coord_type: CoordType, x_clip_type: TextClipType, y_clip_type: TextClipType) -> typing.Sequence[TextRange]: ...

    def get_caret_offset(self) -> builtins.int: ...

    def get_character_at_offset(self, offset: builtins.int) -> builtins.str: ...

    def get_character_count(self) -> builtins.int: ...

    def get_character_extents(self, offset: builtins.int, coords: CoordType) -> typing.Tuple[builtins.int, builtins.int, builtins.int, builtins.int]: ...

    def get_default_attributes(self) -> typing.Sequence[builtins.object]: ...

    def get_n_selections(self) -> builtins.int: ...

    def get_offset_at_point(self, x: builtins.int, y: builtins.int, coords: CoordType) -> builtins.int: ...

    def get_range_extents(self, start_offset: builtins.int, end_offset: builtins.int, coord_type: CoordType) -> TextRectangle: ...

    def get_run_attributes(self, offset: builtins.int) -> typing.Tuple[typing.Sequence[builtins.object], builtins.int, builtins.int]: ...

    def get_selection(self, selection_num: builtins.int) -> typing.Tuple[builtins.str, builtins.int, builtins.int]: ...

    def get_string_at_offset(self, offset: builtins.int, granularity: TextGranularity) -> typing.Tuple[typing.Optional[builtins.str], builtins.int, builtins.int]: ...

    def get_text(self, start_offset: builtins.int, end_offset: builtins.int) -> builtins.str: ...

    def get_text_after_offset(self, offset: builtins.int, boundary_type: TextBoundary) -> typing.Tuple[builtins.str, builtins.int, builtins.int]: ...

    def get_text_at_offset(self, offset: builtins.int, boundary_type: TextBoundary) -> typing.Tuple[builtins.str, builtins.int, builtins.int]: ...

    def get_text_before_offset(self, offset: builtins.int, boundary_type: TextBoundary) -> typing.Tuple[builtins.str, builtins.int, builtins.int]: ...

    def remove_selection(self, selection_num: builtins.int) -> builtins.bool: ...

    def scroll_substring_to(self, start_offset: builtins.int, end_offset: builtins.int, type: ScrollType) -> builtins.bool: ...

    def scroll_substring_to_point(self, start_offset: builtins.int, end_offset: builtins.int, coords: CoordType, x: builtins.int, y: builtins.int) -> builtins.bool: ...

    def set_caret_offset(self, offset: builtins.int) -> builtins.bool: ...

    def set_selection(self, selection_num: builtins.int, start_offset: builtins.int, end_offset: builtins.int) -> builtins.bool: ...

    def do_add_selection(self, start_offset: builtins.int, end_offset: builtins.int) -> builtins.bool: ...

    def do_get_bounded_ranges(self, rect: TextRectangle, coord_type: CoordType, x_clip_type: TextClipType, y_clip_type: TextClipType) -> typing.Sequence[TextRange]: ...

    def do_get_caret_offset(self) -> builtins.int: ...

    def do_get_character_at_offset(self, offset: builtins.int) -> builtins.str: ...

    def do_get_character_count(self) -> builtins.int: ...

    def do_get_character_extents(self, offset: builtins.int, coords: CoordType) -> typing.Tuple[builtins.int, builtins.int, builtins.int, builtins.int]: ...

    def do_get_default_attributes(self) -> typing.Sequence[builtins.object]: ...

    def do_get_n_selections(self) -> builtins.int: ...

    def do_get_offset_at_point(self, x: builtins.int, y: builtins.int, coords: CoordType) -> builtins.int: ...

    def do_get_range_extents(self, start_offset: builtins.int, end_offset: builtins.int, coord_type: CoordType) -> TextRectangle: ...

    def do_get_run_attributes(self, offset: builtins.int) -> typing.Tuple[typing.Sequence[builtins.object], builtins.int, builtins.int]: ...

    def do_get_selection(self, selection_num: builtins.int) -> typing.Tuple[builtins.str, builtins.int, builtins.int]: ...

    def do_get_string_at_offset(self, offset: builtins.int, granularity: TextGranularity) -> typing.Tuple[typing.Optional[builtins.str], builtins.int, builtins.int]: ...

    def do_get_text(self, start_offset: builtins.int, end_offset: builtins.int) -> builtins.str: ...

    def do_get_text_after_offset(self, offset: builtins.int, boundary_type: TextBoundary) -> typing.Tuple[builtins.str, builtins.int, builtins.int]: ...

    def do_get_text_at_offset(self, offset: builtins.int, boundary_type: TextBoundary) -> typing.Tuple[builtins.str, builtins.int, builtins.int]: ...

    def do_get_text_before_offset(self, offset: builtins.int, boundary_type: TextBoundary) -> typing.Tuple[builtins.str, builtins.int, builtins.int]: ...

    def do_remove_selection(self, selection_num: builtins.int) -> builtins.bool: ...

    def do_scroll_substring_to(self, start_offset: builtins.int, end_offset: builtins.int, type: ScrollType) -> builtins.bool: ...

    def do_scroll_substring_to_point(self, start_offset: builtins.int, end_offset: builtins.int, coords: CoordType, x: builtins.int, y: builtins.int) -> builtins.bool: ...

    def do_set_caret_offset(self, offset: builtins.int) -> builtins.bool: ...

    def do_set_selection(self, selection_num: builtins.int, start_offset: builtins.int, end_offset: builtins.int) -> builtins.bool: ...

    def do_text_attributes_changed(self) -> None: ...

    def do_text_caret_moved(self, location: builtins.int) -> None: ...

    def do_text_changed(self, position: builtins.int, length: builtins.int) -> None: ...

    def do_text_selection_changed(self) -> None: ...


class Util(GObject.Object):
    parent: GObject.Object


class Value(GObject.GInterface):

    def get_current_value(self) -> GObject.Value: ...

    def get_increment(self) -> builtins.float: ...

    def get_maximum_value(self) -> GObject.Value: ...

    def get_minimum_increment(self) -> GObject.Value: ...

    def get_minimum_value(self) -> GObject.Value: ...

    def get_range(self) -> typing.Optional[Range]: ...

    def get_sub_ranges(self) -> typing.Sequence[Range]: ...

    def get_value_and_text(self) -> typing.Tuple[builtins.float, builtins.str]: ...

    def set_current_value(self, value: GObject.Value) -> builtins.bool: ...

    def set_value(self, new_value: builtins.float) -> None: ...

    def do_get_current_value(self) -> GObject.Value: ...

    def do_get_increment(self) -> builtins.float: ...

    def do_get_maximum_value(self) -> GObject.Value: ...

    def do_get_minimum_increment(self) -> GObject.Value: ...

    def do_get_minimum_value(self) -> GObject.Value: ...

    def do_get_range(self) -> typing.Optional[Range]: ...

    def do_get_sub_ranges(self) -> typing.Sequence[Range]: ...

    def do_get_value_and_text(self) -> typing.Tuple[builtins.float, builtins.str]: ...

    def do_set_current_value(self, value: GObject.Value) -> builtins.bool: ...

    def do_set_value(self, new_value: builtins.float) -> None: ...


class Window(GObject.GInterface):
    ...


class Hyperlink(GObject.Object, Action):
    parent: GObject.Object

    def get_end_index(self) -> builtins.int: ...

    def get_n_anchors(self) -> builtins.int: ...

    def get_object(self, i: builtins.int) -> Object: ...

    def get_start_index(self) -> builtins.int: ...

    def get_uri(self, i: builtins.int) -> builtins.str: ...

    def is_inline(self) -> builtins.bool: ...

    def is_selected_link(self) -> builtins.bool: ...

    def is_valid(self) -> builtins.bool: ...

    def do_get_end_index(self) -> builtins.int: ...

    def do_get_n_anchors(self) -> builtins.int: ...

    def do_get_object(self, i: builtins.int) -> Object: ...

    def do_get_start_index(self) -> builtins.int: ...

    def do_get_uri(self, i: builtins.int) -> builtins.str: ...

    def do_is_selected_link(self) -> builtins.bool: ...

    def do_is_valid(self) -> builtins.bool: ...

    def do_link_activated(self) -> None: ...

    def do_link_state(self) -> builtins.int: ...


class GObjectAccessible(Object):
    parent: Object

    @staticmethod
    def for_object(obj: GObject.Object) -> Object: ...

    def get_object(self) -> GObject.Object: ...


class Plug(Object, Component):
    parent: Object

    def get_id(self) -> builtins.str: ...

    @staticmethod
    def new() -> Object: ...

    def set_child(self, child: Object) -> None: ...

    def do_get_object_id(self) -> builtins.str: ...


class Socket(Object, Component):
    embedded_plug_id: builtins.str
    parent: Object

    def embed(self, plug_id: builtins.str) -> None: ...

    def is_occupied(self) -> builtins.bool: ...

    @staticmethod
    def new() -> Object: ...

    def do_embed(self, plug_id: builtins.str) -> None: ...


class NoOpObjectFactory(ObjectFactory):
    parent: ObjectFactory

    @staticmethod
    def new() -> ObjectFactory: ...


class NoOpObject(Object, Action, Component, Document, EditableText, Hypertext, Image, Selection, Table, TableCell, Text, Value, Window):  # type: ignore
    parent: Object

    @staticmethod
    def new(obj: GObject.Object) -> Object: ...


class Attribute():
    name: builtins.str
    value: builtins.str

    @staticmethod
    def set_free(attrib_set: typing.Sequence[builtins.object]) -> None: ...


class Implementor():

    def ref_accessible(self) -> Object: ...


class KeyEventStruct():
    keycode: builtins.int
    keyval: builtins.int
    length: builtins.int
    state: builtins.int
    string: builtins.str
    timestamp: builtins.int
    type: builtins.int


class PropertyValues():
    new_value: GObject.Value
    old_value: GObject.Value
    property_name: builtins.str


class Range():

    def copy(self) -> Range: ...

    def free(self) -> None: ...

    def get_description(self) -> builtins.str: ...

    def get_lower_limit(self) -> builtins.float: ...

    def get_upper_limit(self) -> builtins.float: ...

    @staticmethod
    def new(lower_limit: builtins.float, upper_limit: builtins.float, description: builtins.str) -> Range: ...


class Rectangle():
    height: builtins.int
    width: builtins.int
    x: builtins.int
    y: builtins.int


class TextRange():
    bounds: TextRectangle
    content: builtins.str
    end_offset: builtins.int
    start_offset: builtins.int


class TextRectangle():
    height: builtins.int
    width: builtins.int
    x: builtins.int
    y: builtins.int


class HyperlinkStateFlags(GObject.GFlags, builtins.int):
    INLINE = ...  # type: HyperlinkStateFlags


class CoordType(GObject.GEnum, builtins.int):
    PARENT = ...  # type: CoordType
    SCREEN = ...  # type: CoordType
    WINDOW = ...  # type: CoordType


class KeyEventType(GObject.GEnum, builtins.int):
    LAST_DEFINED = ...  # type: KeyEventType
    PRESS = ...  # type: KeyEventType
    RELEASE = ...  # type: KeyEventType


class Layer(GObject.GEnum, builtins.int):
    BACKGROUND = ...  # type: Layer
    CANVAS = ...  # type: Layer
    INVALID = ...  # type: Layer
    MDI = ...  # type: Layer
    OVERLAY = ...  # type: Layer
    POPUP = ...  # type: Layer
    WIDGET = ...  # type: Layer
    WINDOW = ...  # type: Layer


class RelationType(GObject.GEnum, builtins.int):
    CONTROLLED_BY = ...  # type: RelationType
    CONTROLLER_FOR = ...  # type: RelationType
    DESCRIBED_BY = ...  # type: RelationType
    DESCRIPTION_FOR = ...  # type: RelationType
    DETAILS = ...  # type: RelationType
    DETAILS_FOR = ...  # type: RelationType
    EMBEDDED_BY = ...  # type: RelationType
    EMBEDS = ...  # type: RelationType
    ERROR_FOR = ...  # type: RelationType
    ERROR_MESSAGE = ...  # type: RelationType
    FLOWS_FROM = ...  # type: RelationType
    FLOWS_TO = ...  # type: RelationType
    LABELLED_BY = ...  # type: RelationType
    LABEL_FOR = ...  # type: RelationType
    LAST_DEFINED = ...  # type: RelationType
    MEMBER_OF = ...  # type: RelationType
    NODE_CHILD_OF = ...  # type: RelationType
    NODE_PARENT_OF = ...  # type: RelationType
    NULL = ...  # type: RelationType
    PARENT_WINDOW_OF = ...  # type: RelationType
    POPUP_FOR = ...  # type: RelationType
    SUBWINDOW_OF = ...  # type: RelationType

    @staticmethod
    def for_name(name: builtins.str) -> RelationType: ...

    @staticmethod
    def get_name(type: RelationType) -> builtins.str: ...

    @staticmethod
    def register(name: builtins.str) -> RelationType: ...


class Role(GObject.GEnum, builtins.int):
    ACCELERATOR_LABEL = ...  # type: Role
    ALERT = ...  # type: Role
    ANIMATION = ...  # type: Role
    APPLICATION = ...  # type: Role
    ARROW = ...  # type: Role
    ARTICLE = ...  # type: Role
    AUDIO = ...  # type: Role
    AUTOCOMPLETE = ...  # type: Role
    BLOCK_QUOTE = ...  # type: Role
    CALENDAR = ...  # type: Role
    CANVAS = ...  # type: Role
    CAPTION = ...  # type: Role
    CHART = ...  # type: Role
    CHECK_BOX = ...  # type: Role
    CHECK_MENU_ITEM = ...  # type: Role
    COLOR_CHOOSER = ...  # type: Role
    COLUMN_HEADER = ...  # type: Role
    COMBO_BOX = ...  # type: Role
    COMMENT = ...  # type: Role
    CONTENT_DELETION = ...  # type: Role
    CONTENT_INSERTION = ...  # type: Role
    DATE_EDITOR = ...  # type: Role
    DEFINITION = ...  # type: Role
    DESCRIPTION_LIST = ...  # type: Role
    DESCRIPTION_TERM = ...  # type: Role
    DESCRIPTION_VALUE = ...  # type: Role
    DESKTOP_FRAME = ...  # type: Role
    DESKTOP_ICON = ...  # type: Role
    DIAL = ...  # type: Role
    DIALOG = ...  # type: Role
    DIRECTORY_PANE = ...  # type: Role
    DOCUMENT_EMAIL = ...  # type: Role
    DOCUMENT_FRAME = ...  # type: Role
    DOCUMENT_PRESENTATION = ...  # type: Role
    DOCUMENT_SPREADSHEET = ...  # type: Role
    DOCUMENT_TEXT = ...  # type: Role
    DOCUMENT_WEB = ...  # type: Role
    DRAWING_AREA = ...  # type: Role
    EDIT_BAR = ...  # type: Role
    EMBEDDED = ...  # type: Role
    ENTRY = ...  # type: Role
    FILE_CHOOSER = ...  # type: Role
    FILLER = ...  # type: Role
    FONT_CHOOSER = ...  # type: Role
    FOOTER = ...  # type: Role
    FOOTNOTE = ...  # type: Role
    FORM = ...  # type: Role
    FRAME = ...  # type: Role
    GLASS_PANE = ...  # type: Role
    GROUPING = ...  # type: Role
    HEADER = ...  # type: Role
    HEADING = ...  # type: Role
    HTML_CONTAINER = ...  # type: Role
    ICON = ...  # type: Role
    IMAGE = ...  # type: Role
    IMAGE_MAP = ...  # type: Role
    INFO_BAR = ...  # type: Role
    INPUT_METHOD_WINDOW = ...  # type: Role
    INTERNAL_FRAME = ...  # type: Role
    INVALID = ...  # type: Role
    LABEL = ...  # type: Role
    LANDMARK = ...  # type: Role
    LAST_DEFINED = ...  # type: Role
    LAYERED_PANE = ...  # type: Role
    LEVEL_BAR = ...  # type: Role
    LINK = ...  # type: Role
    LIST = ...  # type: Role
    LIST_BOX = ...  # type: Role
    LIST_ITEM = ...  # type: Role
    LOG = ...  # type: Role
    MARK = ...  # type: Role
    MARQUEE = ...  # type: Role
    MATH = ...  # type: Role
    MATH_FRACTION = ...  # type: Role
    MATH_ROOT = ...  # type: Role
    MENU = ...  # type: Role
    MENU_BAR = ...  # type: Role
    MENU_ITEM = ...  # type: Role
    NOTIFICATION = ...  # type: Role
    OPTION_PANE = ...  # type: Role
    PAGE = ...  # type: Role
    PAGE_TAB = ...  # type: Role
    PAGE_TAB_LIST = ...  # type: Role
    PANEL = ...  # type: Role
    PARAGRAPH = ...  # type: Role
    PASSWORD_TEXT = ...  # type: Role
    POPUP_MENU = ...  # type: Role
    PROGRESS_BAR = ...  # type: Role
    PUSH_BUTTON = ...  # type: Role
    RADIO_BUTTON = ...  # type: Role
    RADIO_MENU_ITEM = ...  # type: Role
    RATING = ...  # type: Role
    REDUNDANT_OBJECT = ...  # type: Role
    ROOT_PANE = ...  # type: Role
    ROW_HEADER = ...  # type: Role
    RULER = ...  # type: Role
    SCROLL_BAR = ...  # type: Role
    SCROLL_PANE = ...  # type: Role
    SECTION = ...  # type: Role
    SEPARATOR = ...  # type: Role
    SLIDER = ...  # type: Role
    SPIN_BUTTON = ...  # type: Role
    SPLIT_PANE = ...  # type: Role
    STATIC = ...  # type: Role
    STATUSBAR = ...  # type: Role
    SUBSCRIPT = ...  # type: Role
    SUGGESTION = ...  # type: Role
    SUPERSCRIPT = ...  # type: Role
    TABLE = ...  # type: Role
    TABLE_CELL = ...  # type: Role
    TABLE_COLUMN_HEADER = ...  # type: Role
    TABLE_ROW = ...  # type: Role
    TABLE_ROW_HEADER = ...  # type: Role
    TEAR_OFF_MENU_ITEM = ...  # type: Role
    TERMINAL = ...  # type: Role
    TEXT = ...  # type: Role
    TIMER = ...  # type: Role
    TITLE_BAR = ...  # type: Role
    TOGGLE_BUTTON = ...  # type: Role
    TOOL_BAR = ...  # type: Role
    TOOL_TIP = ...  # type: Role
    TREE = ...  # type: Role
    TREE_ITEM = ...  # type: Role
    TREE_TABLE = ...  # type: Role
    UNKNOWN = ...  # type: Role
    VIDEO = ...  # type: Role
    VIEWPORT = ...  # type: Role
    WINDOW = ...  # type: Role

    @staticmethod
    def for_name(name: builtins.str) -> Role: ...

    @staticmethod
    def get_localized_name(role: Role) -> builtins.str: ...

    @staticmethod
    def get_name(role: Role) -> builtins.str: ...

    @staticmethod
    def register(name: builtins.str) -> Role: ...


class ScrollType(GObject.GEnum, builtins.int):
    ANYWHERE = ...  # type: ScrollType
    BOTTOM_EDGE = ...  # type: ScrollType
    BOTTOM_RIGHT = ...  # type: ScrollType
    LEFT_EDGE = ...  # type: ScrollType
    RIGHT_EDGE = ...  # type: ScrollType
    TOP_EDGE = ...  # type: ScrollType
    TOP_LEFT = ...  # type: ScrollType


class StateType(GObject.GEnum, builtins.int):
    ACTIVE = ...  # type: StateType
    ANIMATED = ...  # type: StateType
    ARMED = ...  # type: StateType
    BUSY = ...  # type: StateType
    CHECKABLE = ...  # type: StateType
    CHECKED = ...  # type: StateType
    DEFAULT = ...  # type: StateType
    DEFUNCT = ...  # type: StateType
    EDITABLE = ...  # type: StateType
    ENABLED = ...  # type: StateType
    EXPANDABLE = ...  # type: StateType
    EXPANDED = ...  # type: StateType
    FOCUSABLE = ...  # type: StateType
    FOCUSED = ...  # type: StateType
    HAS_POPUP = ...  # type: StateType
    HAS_TOOLTIP = ...  # type: StateType
    HORIZONTAL = ...  # type: StateType
    ICONIFIED = ...  # type: StateType
    INDETERMINATE = ...  # type: StateType
    INVALID = ...  # type: StateType
    INVALID_ENTRY = ...  # type: StateType
    LAST_DEFINED = ...  # type: StateType
    MANAGES_DESCENDANTS = ...  # type: StateType
    MODAL = ...  # type: StateType
    MULTISELECTABLE = ...  # type: StateType
    MULTI_LINE = ...  # type: StateType
    OPAQUE = ...  # type: StateType
    PRESSED = ...  # type: StateType
    READ_ONLY = ...  # type: StateType
    REQUIRED = ...  # type: StateType
    RESIZABLE = ...  # type: StateType
    SELECTABLE = ...  # type: StateType
    SELECTABLE_TEXT = ...  # type: StateType
    SELECTED = ...  # type: StateType
    SENSITIVE = ...  # type: StateType
    SHOWING = ...  # type: StateType
    SINGLE_LINE = ...  # type: StateType
    STALE = ...  # type: StateType
    SUPPORTS_AUTOCOMPLETION = ...  # type: StateType
    TRANSIENT = ...  # type: StateType
    TRUNCATED = ...  # type: StateType
    VERTICAL = ...  # type: StateType
    VISIBLE = ...  # type: StateType
    VISITED = ...  # type: StateType

    @staticmethod
    def for_name(name: builtins.str) -> StateType: ...

    @staticmethod
    def get_name(type: StateType) -> builtins.str: ...

    @staticmethod
    def register(name: builtins.str) -> StateType: ...


class TextAttribute(GObject.GEnum, builtins.int):
    BG_COLOR = ...  # type: TextAttribute
    BG_FULL_HEIGHT = ...  # type: TextAttribute
    BG_STIPPLE = ...  # type: TextAttribute
    DIRECTION = ...  # type: TextAttribute
    EDITABLE = ...  # type: TextAttribute
    FAMILY_NAME = ...  # type: TextAttribute
    FG_COLOR = ...  # type: TextAttribute
    FG_STIPPLE = ...  # type: TextAttribute
    INDENT = ...  # type: TextAttribute
    INVALID = ...  # type: TextAttribute
    INVISIBLE = ...  # type: TextAttribute
    JUSTIFICATION = ...  # type: TextAttribute
    LANGUAGE = ...  # type: TextAttribute
    LAST_DEFINED = ...  # type: TextAttribute
    LEFT_MARGIN = ...  # type: TextAttribute
    PIXELS_ABOVE_LINES = ...  # type: TextAttribute
    PIXELS_BELOW_LINES = ...  # type: TextAttribute
    PIXELS_INSIDE_WRAP = ...  # type: TextAttribute
    RIGHT_MARGIN = ...  # type: TextAttribute
    RISE = ...  # type: TextAttribute
    SCALE = ...  # type: TextAttribute
    SIZE = ...  # type: TextAttribute
    STRETCH = ...  # type: TextAttribute
    STRIKETHROUGH = ...  # type: TextAttribute
    STYLE = ...  # type: TextAttribute
    TEXT_POSITION = ...  # type: TextAttribute
    UNDERLINE = ...  # type: TextAttribute
    VARIANT = ...  # type: TextAttribute
    WEIGHT = ...  # type: TextAttribute
    WRAP_MODE = ...  # type: TextAttribute

    @staticmethod
    def for_name(name: builtins.str) -> TextAttribute: ...

    @staticmethod
    def get_name(attr: TextAttribute) -> builtins.str: ...

    @staticmethod
    def get_value(attr: TextAttribute, index_: builtins.int) -> typing.Optional[builtins.str]: ...

    @staticmethod
    def register(name: builtins.str) -> TextAttribute: ...


class TextBoundary(GObject.GEnum, builtins.int):
    CHAR = ...  # type: TextBoundary
    LINE_END = ...  # type: TextBoundary
    LINE_START = ...  # type: TextBoundary
    SENTENCE_END = ...  # type: TextBoundary
    SENTENCE_START = ...  # type: TextBoundary
    WORD_END = ...  # type: TextBoundary
    WORD_START = ...  # type: TextBoundary


class TextClipType(GObject.GEnum, builtins.int):
    BOTH = ...  # type: TextClipType
    MAX = ...  # type: TextClipType
    MIN = ...  # type: TextClipType
    NONE = ...  # type: TextClipType


class TextGranularity(GObject.GEnum, builtins.int):
    CHAR = ...  # type: TextGranularity
    LINE = ...  # type: TextGranularity
    PARAGRAPH = ...  # type: TextGranularity
    SENTENCE = ...  # type: TextGranularity
    WORD = ...  # type: TextGranularity


class ValueType(GObject.GEnum, builtins.int):
    ACCEPTABLE = ...  # type: ValueType
    BAD = ...  # type: ValueType
    BEST = ...  # type: ValueType
    GOOD = ...  # type: ValueType
    HIGH = ...  # type: ValueType
    LAST_DEFINED = ...  # type: ValueType
    LOW = ...  # type: ValueType
    MEDIUM = ...  # type: ValueType
    STRONG = ...  # type: ValueType
    VERY_BAD = ...  # type: ValueType
    VERY_GOOD = ...  # type: ValueType
    VERY_HIGH = ...  # type: ValueType
    VERY_LOW = ...  # type: ValueType
    VERY_STRONG = ...  # type: ValueType
    VERY_WEAK = ...  # type: ValueType
    WEAK = ...  # type: ValueType

    @staticmethod
    def get_localized_name(value_type: ValueType) -> builtins.str: ...

    @staticmethod
    def get_name(value_type: ValueType) -> builtins.str: ...


EventListener = typing.Callable[[Object], None]
EventListenerInit = typing.Callable[[], None]
FocusHandler = typing.Callable[[Object, builtins.bool], None]
Function = typing.Callable[[typing.Optional[builtins.object]], builtins.bool]
KeySnoopFunc = typing.Callable[[KeyEventStruct, typing.Optional[builtins.object]], builtins.int]
PropertyChangeHandler = typing.Callable[[Object, PropertyValues], None]


def attribute_set_free(attrib_set: typing.Sequence[builtins.object]) -> None: ...


def focus_tracker_notify(object: Object) -> None: ...


def get_binary_age() -> builtins.int: ...


def get_default_registry() -> Registry: ...


def get_focus_object() -> Object: ...


def get_interface_age() -> builtins.int: ...


def get_major_version() -> builtins.int: ...


def get_micro_version() -> builtins.int: ...


def get_minor_version() -> builtins.int: ...


def get_root() -> Object: ...


def get_toolkit_name() -> builtins.str: ...


def get_toolkit_version() -> builtins.str: ...


def get_version() -> builtins.str: ...


def relation_type_for_name(name: builtins.str) -> RelationType: ...


def relation_type_get_name(type: RelationType) -> builtins.str: ...


def relation_type_register(name: builtins.str) -> RelationType: ...


def remove_focus_tracker(tracker_id: builtins.int) -> None: ...


def remove_global_event_listener(listener_id: builtins.int) -> None: ...


def remove_key_event_listener(listener_id: builtins.int) -> None: ...


def role_for_name(name: builtins.str) -> Role: ...


def role_get_localized_name(role: Role) -> builtins.str: ...


def role_get_name(role: Role) -> builtins.str: ...


def role_register(name: builtins.str) -> Role: ...


def state_type_for_name(name: builtins.str) -> StateType: ...


def state_type_get_name(type: StateType) -> builtins.str: ...


def state_type_register(name: builtins.str) -> StateType: ...


def text_attribute_for_name(name: builtins.str) -> TextAttribute: ...


def text_attribute_get_name(attr: TextAttribute) -> builtins.str: ...


def text_attribute_get_value(attr: TextAttribute, index_: builtins.int) -> typing.Optional[builtins.str]: ...


def text_attribute_register(name: builtins.str) -> TextAttribute: ...


def text_free_ranges(ranges: typing.Sequence[TextRange]) -> None: ...


def value_type_get_localized_name(value_type: ValueType) -> builtins.str: ...


def value_type_get_name(value_type: ValueType) -> builtins.str: ...


BINARY_AGE: builtins.int
INTERFACE_AGE: builtins.int
MAJOR_VERSION: builtins.int
MICRO_VERSION: builtins.int
MINOR_VERSION: builtins.int
VERSION_MIN_REQUIRED: builtins.int



import logging
import weakref
from gettext import gettext as _
from typing import List, TYPE_CHECKING, Dict, Tuple, Any, TypeVar, Type, Optional

from blueman.main.Config import Config


if TYPE_CHECKING:
    from typing_extensions import TypedDict

    # type is actually Type[T] and default is T but this is not supported https://github.com/python/mypy/issues/3863
    class OptionBase(TypedDict):
        type: type
        default: Any

    class Option(OptionBase, total=False):
        name: str
        desc: str
        range: Tuple[int, int]

    class GSettings(TypedDict):
        schema: str
        path: None
else:
    Option = dict


class BasePlugin:
    __depends__: List[str] = []
    __conflicts__: List[str] = []
    __priority__ = 0

    __description__: str = _("Unspecified")
    __author__: str = _("Unspecified")

    __unloadable__ = True
    __autoload__ = True

    __instance__ = None

    __gsettings__: "GSettings"

    __options__: Dict[str, "Option"] = {}

    def __init__(self, *_args: object) -> None:
        if self.__options__:
            self.__config = Config(
                self.__class__.__gsettings__["schema"],
                self.__class__.__gsettings__["path"]
            )

        weakref.finalize(self, self._on_plugin_delete)

    _T = TypeVar("_T", bound="BasePlugin")

    @classmethod
    def get_instance(cls: Type[_T]) -> Optional[_T]:
        return cls.__instance__

    def _on_plugin_delete(self) -> None:
        self.on_delete()
        logging.debug(f"Deleting plugin instance {self}")

    @classmethod
    def is_configurable(cls) -> bool:
        res = map(lambda x: (len(x) > 2), cls.__options__.values())
        return True in res

    def _unload(self) -> None:
        self.on_unload()

        self.__class__.__instance__ = None

    def _load(self) -> None:
        try:
            self.on_load()
            # self.on_manager_state_changed(applet.Manager != None)
            self.__class__.__instance__ = self
        except Exception as e:
            # AppletPlugin.instances.remove(self)
            self.__class__.__instance__ = None
            logging.exception(e)
            raise

    # virtual methods
    def on_load(self) -> None:
        """Do what is neccessary for the plugin to work like add watches or create ui elements"""
        pass

    def on_unload(self) -> None:
        """Tear down any watches and ui elements created in on_load"""
        raise NotImplementedError

    def on_delete(self) -> None:
        """Do cleanup that needs to happen when plugin is deleted"""
        pass

    def get_option(self, key: str) -> Any:
        if key not in self.__class__.__options__:
            raise KeyError("No such option")
        return self.__config[key]

    def set_option(self, key: str, value: Any) -> None:
        if key not in self.__class__.__options__:
            raise KeyError("No such option")
        opt = self.__class__.__options__[key]

        if type(value) == opt["type"]:
            self.__config[key] = value
            self.option_changed(key, value)
        else:
            raise TypeError(f"Wrong type, must be {repr(opt['type'])}")

    def option_changed(self, key: str, value: Any) -> None:
        pass



from gettext import gettext as _
import os
import logging
from typing import Optional, Tuple

from blueman.bluez.Adapter import Adapter
from blueman.gui.DeviceSelectorList import DeviceSelectorList

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class DeviceSelectorWidget(Gtk.Box):
    def __init__(self, adapter_name: Optional[str] = None, orientation: Gtk.Orientation = Gtk.Orientation.VERTICAL,
                 visible: bool = False) -> None:

        super().__init__(orientation=orientation, spacing=1, vexpand=True,
                         width_request=360, height_request=340,
                         name="DeviceSelectorWidget", visible=visible)

        self.List = DeviceSelectorList(adapter_name)
        if self.List.Adapter is not None:
            self.List.display_known_devices()

        sw = Gtk.ScrolledWindow(hscrollbar_policy=Gtk.PolicyType.NEVER,
                                vscrollbar_policy=Gtk.PolicyType.AUTOMATIC,
                                shadow_type=Gtk.ShadowType.IN)
        sw.add(self.List)
        self.pack_start(sw, True, True, 0)

        # Disable overlay scrolling
        if Gtk.get_minor_version() >= 16:
            sw.props.overlay_scrolling = False

        model = Gtk.ListStore(str, str)
        cell = Gtk.CellRendererText()
        self.cb_adapters = Gtk.ComboBox(model=model, visible=True)
        self.cb_adapters.set_tooltip_text(_("Adapter selection"))
        self.cb_adapters.pack_start(cell, True)
        self.cb_adapters.add_attribute(cell, 'text', 0)

        spinner_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6, height_request=8)
        self.spinner = Gtk.Spinner(halign=Gtk.Align.START, hexpand=True, has_tooltip=True,
                                   tooltip_text=_("Discovering…"), margin=6)

        spinner_box.add(self.cb_adapters)
        spinner_box.add(self.spinner)
        self.add(spinner_box)

        self.cb_adapters.connect("changed", self.on_adapter_selected)

        self.List.connect("adapter-changed", self.on_adapter_changed)
        self.List.connect("adapter-added", self.on_adapter_added)
        self.List.connect("adapter-removed", self.on_adapter_removed)
        self.List.connect("adapter-property-changed", self.on_adapter_prop_changed)

        self.update_adapters_list()
        self.show_all()

    def __del__(self) -> None:
        self.List.destroy()
        logging.debug("Deleting widget")

    def on_adapter_prop_changed(self, _devlist: DeviceSelectorList, adapter: Adapter, key_value: Tuple[str, object]
                                ) -> None:
        key, value = key_value
        if key == "Name" or key == "Alias":
            self.update_adapters_list()
        elif key == "Discovering":
            if not value:
                self.spinner.stop()
            else:
                self.spinner.start()

    def on_adapter_added(self, _devlist: DeviceSelectorList, _adapter_path: str) -> None:
        self.update_adapters_list()

    def on_adapter_removed(self, _devlist: DeviceSelectorList, _adapter_path: str) -> None:
        self.update_adapters_list()

    def on_adapter_selected(self, cb_adapters: Gtk.ComboBox) -> None:
        logging.info("selected")
        tree_iter = cb_adapters.get_active_iter()
        if tree_iter:
            adapter_path = cb_adapters.get_model().get_value(tree_iter, 1)
            if self.List.Adapter:
                if self.List.Adapter.get_object_path() != adapter_path:
                    # Stop discovering on previous adapter
                    self.List.Adapter.stop_discovery()
                    self.List.set_adapter(os.path.basename(adapter_path))
                    # Start discovery on selected adapter
                    self.List.Adapter.start_discovery()

    def on_adapter_changed(self, _devlist: DeviceSelectorList, adapter_path: str) -> None:
        logging.info("changed")
        if adapter_path is None:
            self.update_adapters_list()
        else:
            if self.List.Adapter:
                self.List.display_known_devices()

    def update_adapters_list(self) -> None:
        model = self.cb_adapters.get_model()
        assert isinstance(model, Gtk.ListStore)
        model.clear()
        adapters = self.List.manager.get_adapters()
        num = len(adapters)
        if num == 0:
            self.cb_adapters.props.visible = False
            self.List.props.sensitive = False
        elif num == 1:
            self.cb_adapters.props.visible = False
            self.List.props.sensitive = True
        elif num > 1:
            self.List.props.sensitive = True
            self.cb_adapters.props.visible = True
            for adapter in adapters:
                tree_iter = model.append([adapter.get_name(), adapter.get_object_path()])
                if self.List.Adapter and adapter.get_object_path() == self.List.Adapter.get_object_path():
                    self.cb_adapters.set_active_iter(tree_iter)



import inspect
import logging
from typing import Optional, List

from blueman.Service import Service
from blueman.Sdp import ServiceUUID
from blueman.bluez.Device import Device
from blueman.bluez.errors import BluezDBusException
import blueman.services


def get_service(device: Device, uuid: str) -> Optional[Service]:
    for name, cls in inspect.getmembers(blueman.services, inspect.isclass):
        if ServiceUUID(uuid).short_uuid == cls.__svclass_id__:
            svc: Service = cls(device, uuid)
            return svc
    return None


def get_services(device: Device) -> List[Service]:
    try:
        services = (get_service(device, uuid) for uuid in device['UUIDs'])
        return [service for service in services if service]
    except BluezDBusException as e:
        logging.exception(e)
        return []



from gettext import gettext as _
import logging

major_cls = (
    # translators: device class
    _("Miscellaneous"),
    # translators: device class
    _("Computer"),
    # translators: device class
    _("Phone"),
    # translators: device class
    _("Access point"),
    # translators: device class
    _("Audio/video"),
    # translators: device class
    _("Peripheral"),
    # translators: device class
    _("Imaging"),
    # translators: device class
    _("Wearable"),
    # translators: device class
    _("Toy"),
    # translators: device class
    _("Uncategorized")
)

computer_minor_cls = (
    # translators: device class
    _("Uncategorized"),
    # translators: device class
    _("Desktop"),
    # translators: device class
    _("Server"),
    # translators: device class
    _("Laptop"),
    # translators: device class
    _("Handheld"),
    # translators: device class
    _("Palm"),
    # translators: device class
    _("Wearable")
)

phone_minor_cls = (
    # translators: device class
    _("Uncategorized"),
    # translators: device class
    _("Cellular"),
    # translators: device class
    _("Cordless"),
    # translators: device class
    _("Smart phone"),
    # translators: device class
    _("Modem"),
    # translators: device class
    _("Isdn")
)

access_point_minor_cls = (
    # translators: device class
    _("Fully"),
    # translators: device class
    _("1-17 percent"),
    # translators: device class
    _("17-33 percent"),
    # translators: device class
    _("33-50 percent"),
    # translators: device class
    _("50-67 percent"),
    # translators: device class
    _("67-83 percent"),
    # translators: device class
    _("83-99 percent"),
    # translators: device class
    _("Not available")
)

audio_video_minor_cls = (
    # translators: device class
    _("Uncategorized"),
    # translators: device class
    _("Headset"),
    # translators: device class
    _("Handsfree"),
    # translators: device class
    _("Unknown"),
    # translators: device class
    _("Microphone"),
    # translators: device class
    _("Loudspeaker"),
    # translators: device class
    _("Headphones"),
    # translators: device class
    _("Portable audio"),
    # translators: device class
    _("Car audio"),
    # translators: device class
    _("Set-top box"),
    # translators: device class
    _("Hifi audio"),
    # translators: device class
    _("Vcr"),
    # translators: device class
    _("Video camera"),
    # translators: device class
    _("Camcorder"),
    # translators: device class
    _("Video monitor"),
    # translators: device class
    _("Video display and loudspeaker"),
    # translators: device class
    _("Video conferencing"),
    # translators: device class
    _("Unknown"),
    # translators: device class
    _("Gaming/Toy")
)

peripheral_minor_cls = (
    # translators: device class
    _("Uncategorized"),
    # translators: device class
    _("Keyboard"),
    # translators: device class
    _("Pointing"),
    # translators: device class
    _("Combo")
)

imaging_minor_cls = (
    # translators: device class
    _("Display"),
    # translators: device class
    _("Camera"),
    # translators: device class
    _("Scanner"),
    # translators: device class
    _("Printer")
)

wearable_minor_cls = (
    # translators: device class
    _("Wrist watch"),
    # translators: device class
    _("Pager"),
    # translators: device class
    _("Jacket"),
    # translators: device class
    _("Helmet"),
    # translators: device class
    _("Glasses")
)

toy_minor_cls = (
    # translators: device class
    _("Robot"),
    # translators: device class
    _("Vehicle"),
    # translators: device class
    _("Doll"),
    # translators: device class
    _("Controller"),
    # translators: device class
    _("Game")
)

gatt_appearance_categories = {
    0: ('Unknown', {0: _("Unknown")}),
    1: ('Phone', {0: _("Generic Phone")}),
    2: ('Computer', {0: _("Generic Computer")}),
    3: ('Watch', {0: _("Generic Watch"),
                  1: _("Watch: Sports Watch")}),
    4: ('Clock', {0: _("Generic Clock")}),
    5: ('Display', {0: _("Generic Display")}),
    6: ('Remote Control', {0: _("Generic Remote Control")}),
    7: ('Eye-glasses', {0: _("Generic Eye-glasses")}),
    8: ('Tag', {0: _("Generic Tag")}),
    9: ('Keyring', {0: _("Generic Keyring")}),
    10: ('Media Player', {0: _("Generic Media Player")}),
    11: ('Barcode Scanner', {0: _("Generic Barcode Scanner")}),
    12: ('Thermometer', {0: _("Generic Thermometer"),
                         1: _("Thermometer: Ear")}),
    13: ('Heart rate Sensor', {0: _("Generic Heart rate Sensor"),
                               1: _("Heart Rate Sensor: Heart Rate Belt")}),
    14: ('Blood Pressure', {0: _("Generic Blood Pressure"),
                            1: _("Blood Pressure: Arm"),
                            2: _("Blood Pressure: Wrist")}),
    15: ('Human Interface Device (HID)', {0: _("Human Interface Device (HID)"),
                                          1: _("Keyboard"),
                                          2: _("Mouse"),
                                          3: _("Joystick"),
                                          4: _("Gamepad"),
                                          5: _("Digitizer Tablet"),
                                          6: _("Card Reader"),
                                          7: _("Digital Pen"),
                                          8: _("Barcode Scanner")}),
    16: ('Glucose Meter', {0: _("Generic Glucose Meter")}),
    17: ('Running Walking Sensor', {0: _("Generic: Running Walking Sensor"),
                                    1: _("Running Walking Sensor: In-Shoe"),
                                    2: _("Running Walking Sensor: On-Shoe"),
                                    3: _("Running Walking Sensor: On-Hip")}),
    18: ('Cycling', {0: _("Generic: Cycling"),
                     1: _("Cycling: Cycling Computer"),
                     2: _("Cycling: Speed Sensor"),
                     3: _("Cycling: Cadence Sensor"),
                     4: _("Cycling: Power Sensor"),
                     5: _("Cycling: Speed and Cadence Sensor")}),
    # 19 - 48 reserved
    49: ('Pulse Oximeter', {0: _("Generic: Pulse Oximeter"),
                            1: _("Fingertip"),
                            2: _("Wrist Worn")}),
    50: ('Weight Scale', {0: _("Generic: Weight Scale")}),
    51: ('Personal Mobility Device', {0: _("Generic Personal Mobility Device"),
                                      1: _("Powered Wheelchair"),
                                      2: _("Mobility Scooter")}),
    52: ('Continuous Glucose Monitor', {0: _("Generic Continuous Glucose Monitor")}),
    53: ('Insulin Pump', {0: _("Generic Insulin Pump"),
                          1: _("Insulin Pump, durable pump"),
                          4: _("Insulin Pump, patch pump"),
                          8: _("Insulin Pen")}),
    54: ('Medication Delivery', {0: _("Generic Medication Delivery")}),
    # 55 - 80 reserved
    81: ('Outdoor Sports Activity', {0: _("Generic: Outdoor Sports Activity"),
                                     1: _("Location Display Device"),
                                     2: _("Location and Navigation Display Device"),
                                     3: _("Location Pod"),
                                     4: _("Location and Navigation Pod")})
}


def get_major_class(klass: int) -> str:
    index = (klass >> 8) & 0x1F

    if index > 8:
        return major_cls[9]

    return major_cls[index]


def get_minor_class(klass: int) -> str:
    if not klass:
        return _("Unknown")

    i = (klass >> 8) & 0x1F

    if i == 1:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(computer_minor_cls):
            return computer_minor_cls[minor_index]
        else:
            return ""
    elif i == 2:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(phone_minor_cls):
            return phone_minor_cls[minor_index]
        return ""
    elif i == 3:
        minor_index = (klass >> 5) & 0x07
        if minor_index < len(access_point_minor_cls):
            return access_point_minor_cls[minor_index]
        else:
            return ""
    elif i == 4:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(audio_video_minor_cls):
            return audio_video_minor_cls[minor_index]
        else:
            return ""
    elif i == 5:
        minor_index = (klass >> 6) & 0x03
        if minor_index < len(peripheral_minor_cls):
            return peripheral_minor_cls[minor_index]
        else:
            return ""
    elif i == 6:
        return _("Imaging")

    elif i == 7:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(wearable_minor_cls):
            return wearable_minor_cls[minor_index]
        else:
            return ""
    elif i == 8:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(toy_minor_cls):
            return toy_minor_cls[minor_index]
        else:
            return ""

    return _("Unknown")


# First 10 bits is the category, the following 6 bits sub category
def gatt_appearance_to_name(appearance: int) -> str:
    cat = appearance >> 0x6
    subcat = appearance & 0x3f

    if (19 <= cat <= 48) or (55 <= cat <= 80):
        # These ranges are reserved
        logging.debug(f"Reserved category found: {appearance}")
        return gatt_appearance_categories[0][0]
    elif cat > 81:
        logging.warning(f"Invalid catagory found: {appearance}")
        return gatt_appearance_categories[0][0]

    cat_name, subcats = gatt_appearance_categories[cat]

    if subcat not in subcats:
        return cat_name
    else:
        return subcats[subcat]



from typing import Callable, Union, Dict, Type, TYPE_CHECKING

from blueman.bluez.Network import Network as BluezNetwork
from blueman.plugins.MechanismPlugin import MechanismPlugin
from blueman.main.NetConf import NetConf, DnsMasqHandler, DhcpdHandler, UdhcpdHandler

if TYPE_CHECKING:
    from blueman.main.NetConf import DHCPHandler

DHCPDHANDLERS: Dict[str, Type["DHCPHandler"]] = {
    "DnsMasqHandler": DnsMasqHandler,
    "DhcpdHandler": DhcpdHandler,
    "UdhcpdHandler": UdhcpdHandler
}


class Network(MechanismPlugin):
    def on_load(self) -> None:
        self.parent.add_method("DhcpClient", ("s",), "s", self._run_dhcp_client, pass_sender=True, is_async=True)
        self.parent.add_method("EnableNetwork", ("s", "s", "s"), "", self._enable_network, pass_sender=True)
        self.parent.add_method("ReloadNetwork", (), "", self._reload_network, pass_sender=True)
        self.parent.add_method("DisableNetwork", (), "", self._disable_network, pass_sender=True)

    def _run_dhcp_client(self, object_path: str, caller: str, ok: Callable[[str], None],
                         err: Callable[[Union[Exception, int]], None]) -> None:
        self.timer.stop()

        self.confirm_authorization(caller, "org.blueman.dhcp.client")

        from blueman.main.DhcpClient import DhcpClient

        def dh_error(_dh: DhcpClient, num: int) -> None:
            err(num)
            self.timer.resume()

        def dh_connected(_dh: DhcpClient, ip: str) -> None:
            ok(ip)
            self.timer.resume()

        dh = DhcpClient(BluezNetwork(obj_path=object_path)["Interface"])
        dh.connect("error-occurred", dh_error)
        dh.connect("connected", dh_connected)
        try:
            dh.run()
        except Exception as e:
            err(e)

    def _enable_network(self, ip_address: str, netmask: str, dhcp_handler: str, caller: str) -> None:
        self.confirm_authorization(caller, "org.blueman.network.setup")
        nc = NetConf.get_default()
        nc.set_ipv4(ip_address, netmask)
        nc.set_dhcp_handler(DHCPDHANDLERS[dhcp_handler])
        nc.apply_settings()

    def _reload_network(self, caller: str) -> None:
        nc = NetConf.get_default()
        if nc.ip4_address is None or nc.ip4_mask is None:
            nc.ip4_changed = False
            nc.store()
            return

        self.confirm_authorization(caller, "org.blueman.network.setup")
        nc.apply_settings()

    def _disable_network(self, caller: str) -> None:
        self.confirm_authorization(caller, "org.blueman.network.setup")
        nc = NetConf.get_default()
        nc.remove_settings()
        nc.set_ipv4(None, None)
        nc.store()



from typing import List, Callable

import cairo

from blueman.bluez.Device import Device
from blueman.config.AutoConnectConfig import AutoConnectConfig
from blueman.gui.manager.ManagerDeviceMenu import MenuItemsProvider, ManagerDeviceMenu, DeviceMenuItem
from blueman.plugins.ManagerPlugin import ManagerPlugin
from blueman.Functions import create_menuitem
from blueman.main.DBusProxies import AppletService
from blueman.services import get_services

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class Services(ManagerPlugin, MenuItemsProvider):
    def on_load(self) -> None:
        self.icon_theme = Gtk.IconTheme.get_default()

    def _make_x_icon(self, icon_name: str, size: int) -> cairo.ImageSurface:
        assert self.parent.window is not None

        scale = self.parent.window.get_scale_factor()
        window = self.parent.window.get_window()

        target = self.icon_theme.load_surface(icon_name, size, scale, window, Gtk.IconLookupFlags.FORCE_SIZE)
        assert isinstance(target, cairo.ImageSurface)
        bmx = self.icon_theme.load_surface("blueman-x", size, scale, window, Gtk.IconLookupFlags.FORCE_SIZE)
        assert isinstance(bmx, cairo.ImageSurface)

        x = target.get_width() - bmx.get_width()
        y = target.get_height() - bmx.get_height()
        context = cairo.Context(target)
        context.set_source_surface(bmx, x, y)
        context.paint()

        return target

    def on_request_menu_items(self, manager_menu: ManagerDeviceMenu, device: Device) -> List[DeviceMenuItem]:
        items: List[DeviceMenuItem] = []
        appl = AppletService()

        services = get_services(device)

        connectable_services = [service for service in services if service.connectable]
        for service in connectable_services:
            item: Gtk.MenuItem = create_menuitem(service.name, service.icon)
            if service.description:
                item.props.tooltip_text = service.description
            item.connect("activate", lambda _item: manager_menu.connect_service(service.device, service.uuid))
            items.append(DeviceMenuItem(item, DeviceMenuItem.Group.CONNECT, service.priority))
            item.props.sensitive = service.available
            item.show()

        connected_services = [service for service in services if service.connected_instances]
        for service in connected_services:
            for instance in service.connected_instances:
                surface = self._make_x_icon(service.icon, 16)
                item = create_menuitem(instance.name, surface=surface)
                item.connect(
                    "activate",
                    lambda _item: manager_menu.disconnect_service(service.device, service.uuid, instance.port)
                )
                items.append(DeviceMenuItem(item, DeviceMenuItem.Group.DISCONNECT, service.priority + 100))
                item.show()

        if services:
            config = AutoConnectConfig()
            autoconnect_services = set(config["services"])
            for service in services:
                if service.connected_instances or (device.get_object_path(), service.uuid) in autoconnect_services:
                    item = Gtk.CheckMenuItem(label=service.name)
                    config.bind_to_menuitem(item, device, service.uuid)
                    item.show()
                    items.append(DeviceMenuItem(item, DeviceMenuItem.Group.AUTOCONNECT, service.priority))

        for action, priority in set((action, service.priority)
                                    for service in services for action in service.common_actions
                                    if any(plugin in appl.QueryPlugins() for plugin in action.plugins)):
            item = create_menuitem(action.title, action.icon)
            items.append(DeviceMenuItem(item, DeviceMenuItem.Group.ACTIONS, priority + 200))
            item.show()
            item.connect("activate", self._get_activation_handler(action.callback))

        return items

    @staticmethod
    def _get_activation_handler(callback: Callable[[], None]) -> Callable[[Gtk.MenuItem], None]:
        return lambda _: callback()



from gettext import gettext as _
from typing import Optional

from blueman.Functions import launch
from blueman.main.DBusProxies import ManagerService
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.gui.CommonUi import show_about_dialog
from blueman.gui.applet.PluginDialog import PluginDialog

import gi

from blueman.plugins.applet.PowerManager import PowerManager, PowerStateListener

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gtk, Gdk


class StandardItems(AppletPlugin, PowerStateListener):
    __depends__ = ["StatusIcon", "Menu"]
    __unloadable__ = False
    __description__ = _("Adds standard menu items to the status icon menu")
    __author__ = "walmis"

    def on_load(self) -> None:
        self._plugin_window: Optional[Gtk.Window] = None

        self.parent.Plugins.Menu.add(self, 21)

        self.parent.Plugins.Menu.add(self, 31)

        self.send = self.parent.Plugins.Menu.add(self, 40, text=_("Send _Files to Device") + "…",
                                                 icon_name="blueman-send-file", callback=self.on_send)

        self.parent.Plugins.Menu.add(self, 51)

        self.devices = self.parent.Plugins.Menu.add(self, 60, text=_("_Devices") + "…", icon_name="blueman",
                                                    callback=self.on_devices)

        self.adapters = self.parent.Plugins.Menu.add(self, 70, text=_("Adap_ters") + "…", icon_name="blueman-device",
                                                     callback=self.on_adapters)

        self.parent.Plugins.Menu.add(self, 80, text=_("_Local Services") + "…", icon_name="preferences-desktop",
                                     callback=self.on_local_services)

        self.parent.Plugins.Menu.add(self, 81)

        self.parent.Plugins.Menu.add(self, 90, text=_("_Help"), icon_name='help-about', callback=self.on_about)

        self.parent.Plugins.Menu.add(self, 85, text=_("_Plugins"), icon_name="blueman-plugin", callback=self.on_plugins)

        self.parent.Plugins.StatusIcon.connect("activate", lambda _status_icon: self.on_devices())

    def change_sensitivity(self, sensitive: bool) -> None:
        if 'PowerManager' in self.parent.Plugins.get_loaded():
            power = self.parent.Plugins.PowerManager.get_bluetooth_status()
        else:
            power = True

        sensitive = sensitive and self.parent.Manager is not None and power
        self.send.set_sensitive(sensitive)
        self.devices.set_sensitive(sensitive)
        self.adapters.set_sensitive(sensitive)

    def on_manager_state_changed(self, state: bool) -> None:
        self.change_sensitivity(state)

    def on_power_state_changed(self, manager: PowerManager, state: bool) -> None:
        self.change_sensitivity(state)

    def on_send(self) -> None:
        launch("blueman-sendto", name=_("File Sender"))

    def on_devices(self) -> None:
        m = ManagerService()
        m.startstop()

    def on_adapters(self) -> None:
        launch("blueman-adapters", name=_("Adapter Preferences"))

    def on_local_services(self) -> None:
        launch("blueman-services", name=_("Service Preferences"))

    def on_about(self) -> None:
        about = show_about_dialog("Blueman " + _("applet"), run=False)

        im = Gtk.Image(icon_name="blueman-plugin", pixel_size=16)
        button = Gtk.Button(label=_("Plugins"), visible=True, image=im)

        button.connect("clicked", lambda _button: self.on_plugins())

        about.action_area.pack_start(button, True, True, 0)
        about.action_area.reorder_child(button, 0)

        about.run()
        about.destroy()

    def on_plugins(self) -> None:
        def on_close(win: Gtk.Window, _event: Gdk.Event) -> bool:
            win.destroy()
            self._plugin_window = None
            return False

        if self._plugin_window:
            self._plugin_window.present()
        else:
            self._plugin_window = PluginDialog(self.parent)
            self._plugin_window.connect("delete-event", on_close)
            self._plugin_window.show()



from gettext import gettext as _
from typing import Optional, Tuple

from blueman.bluez.Device import Device
from blueman.gui.DeviceList import DeviceList
from blueman.gui.DeviceSelectorWidget import DeviceSelectorWidget

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class DeviceSelectorDialog(Gtk.Dialog):
    def __init__(self, title: str = _("Select Device"), parent: Optional[Gtk.Container] = None, discover: bool = True,
                 adapter_name: Optional[str] = None) -> None:
        super().__init__(title=title, name="DeviceSelectorDialog", parent=parent, icon_name="blueman", resizable=False)
        self.add_buttons('_Cancel', Gtk.ResponseType.REJECT, '_OK', Gtk.ResponseType.ACCEPT)

        self.vbox.props.halign = Gtk.Align.CENTER
        self.vbox.props.valign = Gtk.Align.CENTER
        self.vbox.props.hexpand = True
        self.vbox.props.vexpand = True
        self.vbox.props.margin = 6

        self.selector = DeviceSelectorWidget(adapter_name=adapter_name, visible=True)
        self.vbox.pack_start(self.selector, True, True, 0)

        self.selection: Optional[Tuple[str, Device]] = None

        self.selector.List.connect("device-selected", self.on_device_selected)
        self.selector.List.connect("adapter-changed", self.on_adapter_changed)
        if discover:
            self.selector.List.discover_devices()

        self.selector.List.connect("row-activated", self.on_row_activated)

    def close(self) -> None:
        self.selector.destroy()
        super().close()

    def on_row_activated(self, _treeview: Gtk.TreeView, _path: Gtk.TreePath, _view_column: Gtk.TreeViewColumn,
                         *_args: object) -> None:
        self.response(Gtk.ResponseType.ACCEPT)

    def on_adapter_changed(self, _devlist: DeviceList, _adapter: str) -> None:
        self.selection = None

    def on_device_selected(self, devlist: DeviceList, device: Device, _tree_iter: Gtk.TreeIter) -> None:
        assert devlist.Adapter is not None
        self.selection = (devlist.Adapter.get_object_path(), device)



from gettext import gettext as _
from typing import Callable, Union, TYPE_CHECKING

from _blueman import RFCOMMError
from gi.repository import GLib

from blueman.Service import Service
from blueman.bluez.errors import BluezDBusException
if TYPE_CHECKING:
    from blueman.main.NetworkManager import NMConnectionError
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.bluez.Device import Device
from blueman.services.Functions import get_service

import logging

from blueman.services.meta import SerialService, NetworkService


class RFCOMMConnectedListener:
    def on_rfcomm_connected(self, service: SerialService, port: str) -> None:
        ...

    def on_rfcomm_disconnect(self, port: int) -> None:
        ...


class RFCOMMConnectHandler:
    def rfcomm_connect_handler(self, service: SerialService, reply: Callable[[str], None],
                               err: Callable[[Union[RFCOMMError, GLib.Error]], None]) -> bool:
        ...


class ServiceConnectHandler:
    def service_connect_handler(self, service: Service, ok: Callable[[], None],
                                err: Callable[[Union["NMConnectionError", GLib.Error]], None]) -> bool:
        ...

    def service_disconnect_handler(self, service: Service, ok: Callable[[], None],
                                   err: Callable[[Union["NMConnectionError", GLib.Error]], None]) -> bool:
        ...


class DBusService(AppletPlugin):
    __depends__ = ["StatusIcon"]
    __unloadable__ = False
    __description__ = _("Provides DBus API for other Blueman components")
    __author__ = "Walmis"

    def on_load(self) -> None:
        self._add_dbus_method("QueryPlugins", (), "as", self.parent.Plugins.get_loaded)
        self._add_dbus_method("QueryAvailablePlugins", (), "as", lambda: list(self.parent.Plugins.get_classes()))
        self._add_dbus_method("SetPluginConfig", ("s", "b"), "", self.parent.Plugins.set_config)
        self._add_dbus_method("ConnectService", ("o", "s"), "", self.connect_service, is_async=True)
        self._add_dbus_method("DisconnectService", ("o", "s", "d"), "", self._disconnect_service, is_async=True)
        self._add_dbus_method("OpenPluginDialog", (), "", self._open_plugin_dialog)

    def connect_service(self, object_path: str, uuid: str, ok: Callable[[], None],
                        err: Callable[[Union[BluezDBusException, "NMConnectionError",
                                             RFCOMMError, GLib.Error, str]], None]) -> None:
        try:
            self.parent.Plugins.RecentConns
        except KeyError:
            logging.warning("RecentConns plugin is unavailable")
        else:
            self.parent.Plugins.RecentConns.notify(object_path, uuid)

        if uuid == '00000000-0000-0000-0000-000000000000':
            device = Device(obj_path=object_path)
            device.connect(reply_handler=ok, error_handler=err)
        else:
            service = get_service(Device(obj_path=object_path), uuid)
            assert service is not None

            if any(plugin.service_connect_handler(service, ok, err)
                   for plugin in self.parent.Plugins.get_loaded_plugins(ServiceConnectHandler)):
                pass
            elif isinstance(service, SerialService):
                def reply(rfcomm: str) -> None:
                    assert isinstance(service, SerialService)  # https://github.com/python/mypy/issues/2608
                    for plugin in self.parent.Plugins.get_loaded_plugins(RFCOMMConnectedListener):
                        plugin.on_rfcomm_connected(service, rfcomm)
                    ok()

                if not any(plugin.rfcomm_connect_handler(service, reply, err)
                           for plugin in self.parent.Plugins.get_loaded_plugins(RFCOMMConnectHandler)):
                    service.connect(reply_handler=lambda port: ok(), error_handler=err)
            elif isinstance(service, NetworkService):
                service.connect(reply_handler=lambda interface: ok(), error_handler=err)
            else:
                logging.info("No handler registered")
                err("Service not supported\nPossibly the plugin that handles this service is not loaded")

    def _disconnect_service(self, object_path: str, uuid: str, port: int, ok: Callable[[], None],
                            err: Callable[[Union[BluezDBusException, "NMConnectionError",
                                                 GLib.Error, str]], None]) -> None:
        if uuid == '00000000-0000-0000-0000-000000000000':
            device = Device(obj_path=object_path)
            device.disconnect(reply_handler=ok, error_handler=err)
        else:
            service = get_service(Device(obj_path=object_path), uuid)
            assert service is not None

            if any(plugin.service_disconnect_handler(service, ok, err)
                   for plugin in self.parent.Plugins.get_loaded_plugins(ServiceConnectHandler)):
                pass
            elif isinstance(service, SerialService):
                service.disconnect(port, reply_handler=ok, error_handler=err)

                for plugin in self.parent.Plugins.get_loaded_plugins(RFCOMMConnectedListener):
                    plugin.on_rfcomm_disconnect(port)

                logging.info("Disconnecting rfcomm device")
            elif isinstance(service, NetworkService):
                service.disconnect(reply_handler=ok, error_handler=err)

    def _open_plugin_dialog(self) -> None:
        self.parent.Plugins.StandardItems.on_plugins()



import os
import struct
from blueman.plugins.MechanismPlugin import MechanismPlugin
from blueman.plugins.applet.KillSwitch import RFKILL_TYPE_BLUETOOTH, RFKILL_OP_CHANGE_ALL

if not os.path.exists('/dev/rfkill'):
    raise ImportError("Hardware kill switch not found")


class RfKill(MechanismPlugin):
    def on_load(self) -> None:
        self.parent.add_method("SetRfkillState", ("b",), "", self._set_rfkill_state, pass_sender=True)

    def _set_rfkill_state(self, state: bool, caller: str) -> None:
        self.confirm_authorization(caller, "org.blueman.rfkill.setstate")
        with open('/dev/rfkill', 'r+b', buffering=0) as f:
            f.write(struct.pack("IBBBB", 0, RFKILL_TYPE_BLUETOOTH, RFKILL_OP_CHANGE_ALL, (0 if state else 1), 0))



import builtins
import typing

from gi.repository import GObject


class Context():
    ...


class Device():
    ...


class FontFace():
    ...


class FontOptions():
    ...


class Matrix():
    ...


class Path():
    ...


class Pattern():
    ...


class Rectangle():
    height: builtins.float
    width: builtins.float
    x: builtins.float
    y: builtins.float


class RectangleInt():
    height: builtins.int
    width: builtins.int
    x: builtins.int
    y: builtins.int


class Region():
    ...


class ScaledFont():
    ...


class Surface():
    ...


class Antialias(GObject.GEnum, builtins.int):
    BEST = ...  # type: Antialias
    DEFAULT = ...  # type: Antialias
    FAST = ...  # type: Antialias
    GOOD = ...  # type: Antialias
    GRAY = ...  # type: Antialias
    NONE = ...  # type: Antialias
    SUBPIXEL = ...  # type: Antialias


class Content(GObject.GEnum, builtins.int):
    ALPHA = ...  # type: Content
    COLOR = ...  # type: Content
    COLOR_ALPHA = ...  # type: Content


class DeviceType(GObject.GEnum, builtins.int):
    COGL = ...  # type: DeviceType
    DRM = ...  # type: DeviceType
    GL = ...  # type: DeviceType
    INVALID = ...  # type: DeviceType
    SCRIPT = ...  # type: DeviceType
    WIN32 = ...  # type: DeviceType
    XCB = ...  # type: DeviceType
    XLIB = ...  # type: DeviceType
    XML = ...  # type: DeviceType


class Extend(GObject.GEnum, builtins.int):
    NONE = ...  # type: Extend
    PAD = ...  # type: Extend
    REFLECT = ...  # type: Extend
    REPEAT = ...  # type: Extend


class FillRule(GObject.GEnum, builtins.int):
    EVEN_ODD = ...  # type: FillRule
    WINDING = ...  # type: FillRule


class Filter(GObject.GEnum, builtins.int):
    BEST = ...  # type: Filter
    BILINEAR = ...  # type: Filter
    FAST = ...  # type: Filter
    GAUSSIAN = ...  # type: Filter
    GOOD = ...  # type: Filter
    NEAREST = ...  # type: Filter


class FontSlant(GObject.GEnum, builtins.int):
    ITALIC = ...  # type: FontSlant
    NORMAL = ...  # type: FontSlant
    OBLIQUE = ...  # type: FontSlant


class FontType(GObject.GEnum, builtins.int):
    FT = ...  # type: FontType
    QUARTZ = ...  # type: FontType
    TOY = ...  # type: FontType
    USER = ...  # type: FontType
    WIN32 = ...  # type: FontType


class FontWeight(GObject.GEnum, builtins.int):
    BOLD = ...  # type: FontWeight
    NORMAL = ...  # type: FontWeight


class Format(GObject.GEnum, builtins.int):
    A1 = ...  # type: Format
    A8 = ...  # type: Format
    ARGB32 = ...  # type: Format
    INVALID = ...  # type: Format
    RGB16_565 = ...  # type: Format
    RGB24 = ...  # type: Format
    RGB30 = ...  # type: Format


class HintMetrics(GObject.GEnum, builtins.int):
    DEFAULT = ...  # type: HintMetrics
    OFF = ...  # type: HintMetrics
    ON = ...  # type: HintMetrics


class HintStyle(GObject.GEnum, builtins.int):
    DEFAULT = ...  # type: HintStyle
    FULL = ...  # type: HintStyle
    MEDIUM = ...  # type: HintStyle
    NONE = ...  # type: HintStyle
    SLIGHT = ...  # type: HintStyle


class LineCap(GObject.GEnum, builtins.int):
    BUTT = ...  # type: LineCap
    ROUND = ...  # type: LineCap
    SQUARE = ...  # type: LineCap


class LineJoin(GObject.GEnum, builtins.int):
    BEVEL = ...  # type: LineJoin
    MITER = ...  # type: LineJoin
    ROUND = ...  # type: LineJoin


class Operator(GObject.GEnum, builtins.int):
    ADD = ...  # type: Operator
    ATOP = ...  # type: Operator
    CLEAR = ...  # type: Operator
    COLOR_BURN = ...  # type: Operator
    COLOR_DODGE = ...  # type: Operator
    DARKEN = ...  # type: Operator
    DEST = ...  # type: Operator
    DEST_ATOP = ...  # type: Operator
    DEST_IN = ...  # type: Operator
    DEST_OUT = ...  # type: Operator
    DEST_OVER = ...  # type: Operator
    DIFFERENCE = ...  # type: Operator
    EXCLUSION = ...  # type: Operator
    HARD_LIGHT = ...  # type: Operator
    HSL_COLOR = ...  # type: Operator
    HSL_HUE = ...  # type: Operator
    HSL_LUMINOSITY = ...  # type: Operator
    HSL_SATURATION = ...  # type: Operator
    IN = ...  # type: Operator
    LIGHTEN = ...  # type: Operator
    MULTIPLY = ...  # type: Operator
    OUT = ...  # type: Operator
    OVER = ...  # type: Operator
    OVERLAY = ...  # type: Operator
    SATURATE = ...  # type: Operator
    SCREEN = ...  # type: Operator
    SOFT_LIGHT = ...  # type: Operator
    SOURCE = ...  # type: Operator
    XOR = ...  # type: Operator


class PathDataType(GObject.GEnum, builtins.int):
    CLOSE_PATH = ...  # type: PathDataType
    CURVE_TO = ...  # type: PathDataType
    LINE_TO = ...  # type: PathDataType
    MOVE_TO = ...  # type: PathDataType


class PatternType(GObject.GEnum, builtins.int):
    LINEAR = ...  # type: PatternType
    MESH = ...  # type: PatternType
    RADIAL = ...  # type: PatternType
    RASTER_SOURCE = ...  # type: PatternType
    SOLID = ...  # type: PatternType
    SURFACE = ...  # type: PatternType


class RegionOverlap(GObject.GEnum, builtins.int):
    IN = ...  # type: RegionOverlap
    OUT = ...  # type: RegionOverlap
    PART = ...  # type: RegionOverlap


class Status(GObject.GEnum, builtins.int):
    CLIP_NOT_REPRESENTABLE = ...  # type: Status
    DEVICE_ERROR = ...  # type: Status
    DEVICE_FINISHED = ...  # type: Status
    DEVICE_TYPE_MISMATCH = ...  # type: Status
    FILE_NOT_FOUND = ...  # type: Status
    FONT_TYPE_MISMATCH = ...  # type: Status
    INVALID_CLUSTERS = ...  # type: Status
    INVALID_CONTENT = ...  # type: Status
    INVALID_DASH = ...  # type: Status
    INVALID_DSC_COMMENT = ...  # type: Status
    INVALID_FORMAT = ...  # type: Status
    INVALID_INDEX = ...  # type: Status
    INVALID_MATRIX = ...  # type: Status
    INVALID_MESH_CONSTRUCTION = ...  # type: Status
    INVALID_PATH_DATA = ...  # type: Status
    INVALID_POP_GROUP = ...  # type: Status
    INVALID_RESTORE = ...  # type: Status
    INVALID_SIZE = ...  # type: Status
    INVALID_SLANT = ...  # type: Status
    INVALID_STATUS = ...  # type: Status
    INVALID_STRIDE = ...  # type: Status
    INVALID_STRING = ...  # type: Status
    INVALID_VISUAL = ...  # type: Status
    INVALID_WEIGHT = ...  # type: Status
    JBIG2_GLOBAL_MISSING = ...  # type: Status
    NEGATIVE_COUNT = ...  # type: Status
    NO_CURRENT_POINT = ...  # type: Status
    NO_MEMORY = ...  # type: Status
    NULL_POINTER = ...  # type: Status
    PATTERN_TYPE_MISMATCH = ...  # type: Status
    READ_ERROR = ...  # type: Status
    SUCCESS = ...  # type: Status
    SURFACE_FINISHED = ...  # type: Status
    SURFACE_TYPE_MISMATCH = ...  # type: Status
    TEMP_FILE_ERROR = ...  # type: Status
    USER_FONT_ERROR = ...  # type: Status
    USER_FONT_IMMUTABLE = ...  # type: Status
    USER_FONT_NOT_IMPLEMENTED = ...  # type: Status
    WRITE_ERROR = ...  # type: Status


class SubpixelOrder(GObject.GEnum, builtins.int):
    BGR = ...  # type: SubpixelOrder
    DEFAULT = ...  # type: SubpixelOrder
    RGB = ...  # type: SubpixelOrder
    VBGR = ...  # type: SubpixelOrder
    VRGB = ...  # type: SubpixelOrder


class SurfaceType(GObject.GEnum, builtins.int):
    BEOS = ...  # type: SurfaceType
    COGL = ...  # type: SurfaceType
    DIRECTFB = ...  # type: SurfaceType
    DRM = ...  # type: SurfaceType
    GL = ...  # type: SurfaceType
    GLITZ = ...  # type: SurfaceType
    IMAGE = ...  # type: SurfaceType
    OS2 = ...  # type: SurfaceType
    PDF = ...  # type: SurfaceType
    PS = ...  # type: SurfaceType
    QT = ...  # type: SurfaceType
    QUARTZ = ...  # type: SurfaceType
    QUARTZ_IMAGE = ...  # type: SurfaceType
    RECORDING = ...  # type: SurfaceType
    SCRIPT = ...  # type: SurfaceType
    SKIA = ...  # type: SurfaceType
    SUBSURFACE = ...  # type: SurfaceType
    SVG = ...  # type: SurfaceType
    TEE = ...  # type: SurfaceType
    VG = ...  # type: SurfaceType
    WIN32 = ...  # type: SurfaceType
    WIN32_PRINTING = ...  # type: SurfaceType
    XCB = ...  # type: SurfaceType
    XLIB = ...  # type: SurfaceType
    XML = ...  # type: SurfaceType


class TextClusterFlags(GObject.GEnum, builtins.int):
    BACKWARD = ...  # type: TextClusterFlags


def image_surface_create() -> None: ...





import errno
import logging
import os
import re
import subprocess
import termios
import tty
from typing import Any, List, Iterable, Callable, Optional, Tuple, Union, MutableSequence, IO

from gi.repository import GObject
from gi.repository import GLib

from blueman.Functions import open_rfcomm

from blueman.bluemantyping import GSignals

pppd_errors = {
    1: "An immediately fatal error of some kind  occurred, such as an essential system call failing, "
       "or running out of virtual memory.",
    2: "An  error  was detected in processing the options given, such as two mutually exclusive options being used.",
    3: "Pppd is not setuid-root and the invoking user is not root.",
    4: "The kernel does not support PPP, for example, the  PPP kernel driver is not included or cannot be loaded.",
    5: "Pppd terminated because it was sent a SIGINT, SIGTERM or SIGHUP signal.",
    6: "The serial port could not be locked.",
    7: "The serial port could not be opened.",
    8: "The connect script failed (returned a non-zero exit status).",
    9: "The command specified as the argument to the  pty  option  could not be run.",
    10: "The PPP negotiation failed, that is, it didn't reach the point where at least one network protocol "
        "(e.g. IP) was running.",
    11: "The peer system failed (or refused) to authenticate itself.",
    12: "The link was established successfully and terminated because  it was idle.",
    13: "The link was established successfully and terminated because the connect time limit was reached.",
    14: "Callback was negotiated  and  an  incoming  call  should  arrive shortly.",
    15: "The link was terminated because the peer is not responding to echo requests.",
    16: "The link was terminated by the modem hanging up.",
    17: "The PPP negotiation failed because serial loopback was detected.",
    18: "The init script failed (returned a non-zero exit status).",
    19: "We failed to authenticate ourselves to the peer."
}


class PPPException(Exception):
    pass


class PPPConnection(GObject.GObject):
    __gsignals__: GSignals = {  # arg: interface name eg. ppp0
        'connected': (GObject.SignalFlags.NO_HOOKS, None, (str,)),
        'error-occurred': (GObject.SignalFlags.NO_HOOKS, None, (str,))
    }

    def __init__(self, port: str, number: str = "*99#", apn: str = "", user: str = "", pwd: str = "") -> None:
        super().__init__()

        self.apn = apn
        self.number = number
        self.user = user
        self.pwd = pwd
        self.port = port
        self.interface: Optional[str] = None

        self.commands: MutableSequence[Union[str, Tuple[str, Callable[[List[str]], None], Iterable[str]]]] = [
            "ATZ E0 V1 X4 &C1 +FCLASS=0",
            "ATE0",
            "AT+GCAP",
            (
                f"ATD{self.number}",
                self.connect_callback,
                ["CONNECT", "NO CARRIER", "BUSY", "NO ANSWER", "NO DIALTONE", "OK", "ERROR"]
            )
        ]
        if self.apn != "":
            self.commands.insert(-1, f'AT+CGDCONT=1,"IP","{self.apn}"')

    def cleanup(self) -> None:
        os.close(self.file)

    def connect_callback(self, response: List[str]) -> None:
        if "CONNECT" in response:
            logging.info("Starting pppd")
            self.pppd = subprocess.Popen(
                ["/usr/sbin/pppd", f"{self.port}", "115200", "defaultroute", "updetach", "usepeerdns"], bufsize=1,
                stdout=subprocess.PIPE)
            assert self.pppd.stdout is not None
            GLib.io_add_watch(self.pppd.stdout, GLib.IO_IN | GLib.IO_ERR | GLib.IO_HUP, self.on_pppd_stdout)
            GLib.timeout_add(1000, self.check_pppd)

            self.cleanup()
        else:
            self.cleanup()
            raise PPPException(f"Bad modem response {response[0]}, expected CONNECT")

    def __cmd_response_cb(self, response: Optional[List[str]], exception: Optional[PPPException],
                          command_id: int) -> None:
        if exception:
            self.emit("error-occurred", str(exception))
        else:
            command = self.commands[command_id]
            if isinstance(command, tuple):
                assert response is not None
                try:
                    command[1](response)
                except PPPException as e:
                    self.emit("error-occurred", str(e))
                    return

            self.send_commands(command_id + 1)

    def send_commands(self, start: int = 0) -> None:
        try:
            item = self.commands[start]
        except IndexError:
            return

        self.send_command(item[0] if isinstance(item, tuple) else item)
        self.wait_for_reply(start)

    def connect_rfcomm(self) -> None:

        self.file = open_rfcomm(self.port, os.O_RDWR)

        tty.setraw(self.file)

        attrs: List[Any] = termios.tcgetattr(self.file)

        attrs[0] &= ~(termios.IGNCR | termios.ICRNL | termios.IUCLC | termios.INPCK | termios.IXON | termios.IXANY |
                      termios.IGNPAR)
        attrs[1] &= ~(termios.OPOST | termios.OLCUC | termios.OCRNL | termios.ONLCR | termios.ONLRET)
        attrs[3] &= ~(termios.ICANON | getattr(termios, 'XCASE', 4) | termios.ECHO | termios.ECHOE | termios.ECHONL)
        attrs[3] &= ~(termios.ECHO | termios.ECHOE)
        attrs[6][termios.VMIN] = 1
        attrs[6][termios.VTIME] = 0
        attrs[6][termios.VEOF] = 1

        attrs[2] &= ~(termios.CBAUD | termios.CSIZE | termios.CSTOPB | termios.CLOCAL | termios.PARENB)
        attrs[2] |= (termios.B9600 | termios.CS8 | termios.CREAD | termios.PARENB)

        termios.tcsetattr(self.file, termios.TCSANOW, attrs)

        termios.tcflush(self.file, termios.TCIOFLUSH)

        self.send_commands()

    def on_pppd_stdout(self, source: IO[bytes], cond: GLib.IOCondition) -> bool:
        if cond & GLib.IO_ERR or cond & GLib.IO_HUP:
            return False

        line = source.readline().decode('utf-8')
        m = re.match(r'Using interface (ppp[0-9]*)', line)
        if m:
            self.interface = m.groups()[0]

        logging.info(line)

        return True

    def check_pppd(self) -> bool:
        status = self.pppd.poll()
        if status is not None:
            if status == 0:
                self.emit("connected", self.interface)
            else:
                try:
                    msg = "pppd exited: " + pppd_errors[int(status)]
                except KeyError:
                    msg = "pppd exited with unknown error"

                self.emit("error-occurred", msg)

            logging.warning("pppd exited with status %d" % status)
            return False
        return True

    def send_command(self, command: str) -> None:
        logging.info(f"--> {command}")
        out = f"{command}\r\n"
        os.write(self.file, out.encode("UTF-8"))
        termios.tcdrain(self.file)

    def on_data_ready(self, _source: int, condition: GLib.IOCondition, command_id: int) -> bool:
        if condition & GLib.IO_ERR or condition & GLib.IO_HUP:
            GLib.source_remove(self.timeout)
            self.__cmd_response_cb(None, PPPException("Socket error"), command_id)
            self.cleanup()
            return False
        try:
            self.buffer += os.read(self.file, 1).decode('utf-8')
        except OSError as e:
            if e.errno == errno.EAGAIN:
                logging.error("Got EAGAIN")
                return True
            else:
                self.__cmd_response_cb(None, PPPException("Socket error"), command_id)
                logging.exception(e)
                self.cleanup()
                return False

        terminators = self.commands[command_id][2] if isinstance(self.commands[command_id], tuple) else ["OK", "ERROR"]

        lines = self.buffer.split("\r\n")

        if any(terminator in line for line in lines for terminator in terminators):
            lines = [x.strip("\r\n") for x in lines if x != ""]
            logging.info(f"<-- {lines}")

            self.__cmd_response_cb(lines, None, command_id)
            return False

        return True

    def wait_for_reply(self, command_id: int) -> None:
        def on_timeout() -> bool:
            GLib.source_remove(self.io_watch)
            self.__cmd_response_cb(None, PPPException("Modem initialization timed out"), command_id)
            self.cleanup()
            return False

        self.buffer = ""
        self.term_found = False

        self.io_watch = GLib.io_add_watch(self.file, GLib.IO_IN | GLib.IO_ERR | GLib.IO_HUP, self.on_data_ready,
                                          command_id)
        self.timeout = GLib.timeout_add(15000, on_timeout)



from unittest import TestCase

from blueman.bluez.obex.Manager import Manager
from blueman.gobject import SingletonGObjectMeta


class TestManager(TestCase):
    def test_metaclass(self):
        self.assertIsInstance(Manager, SingletonGObjectMeta)



from gettext import gettext as _
import os
import logging
import traceback
import importlib
from types import ModuleType
from typing import Dict, List, Type, TypeVar, Iterable, Optional, Any

from gi.repository import GObject, Gio

from blueman.Functions import bmexit
from blueman.gui.CommonUi import ErrorDialog
from blueman.main.Config import Config
from blueman.plugins.BasePlugin import BasePlugin
from blueman.bluemantyping import GSignals


class LoadException(Exception):
    pass


_T = TypeVar("_T", bound=BasePlugin)


class PluginManager(GObject.GObject):
    __gsignals__: GSignals = {
        'plugin-loaded': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_STRING,)),
        'plugin-unloaded': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_STRING,)),
    }

    def __init__(self, plugin_class: Type[_T], module_path: ModuleType, parent: object) -> None:
        super().__init__()
        self.__deps: Dict[str, List[str]] = {}
        self.__cfls: Dict[str, List[str]] = {}
        self.__plugins: Dict[str, _T] = {}
        self.__classes: Dict[str, Type[_T]] = {}
        self.__loaded: List[str] = []
        self.parent = parent

        self.module_path = module_path
        self.plugin_class = plugin_class

    @property
    def config_list(self) -> List[str]:
        return []

    def get_classes(self) -> Dict[str, Type[_T]]:
        return self.__classes

    def get_loaded(self) -> List[str]:
        return self.__loaded

    def get_dependencies(self) -> Dict[str, List[str]]:
        return self.__deps

    def get_conflicts(self) -> Dict[str, List[str]]:
        return self.__cfls

    def load_plugin(self, name: Optional[str] = None, user_action: bool = False) -> None:
        if name:
            try:
                self.__load_plugin(self.__classes[name])
            except LoadException:
                pass
            except Exception:
                if user_action:
                    d = ErrorDialog(_("<b>An error has occured while loading "
                                      "a plugin. Please notify the developers "
                                      "with the content of this message to our </b>\n"
                                      "<a href=\"http://github.com/blueman-project/blueman/issues\">website.</a>"),
                                    excp=traceback.format_exc())
                    d.run()
                    d.destroy()
                    raise

            return

        path = os.path.dirname(self.module_path.__file__)
        plugins = []
        for root, dirs, files in os.walk(path):
            for f in files:
                if f.endswith(".py") and not (f.endswith(".pyc") or f.endswith("_.py")):
                    plugins.append(f[0:-3])

        logging.info(plugins)
        for plugin in plugins:
            try:
                importlib.import_module(self.module_path.__name__ + f".{plugin}")
            except ImportError:
                logging.error(f"Unable to load plugin module {plugin}", exc_info=True)

        for cls in self.plugin_class.__subclasses__():
            self.__classes[cls.__name__] = cls
            if cls.__name__ not in self.__deps:
                self.__deps[cls.__name__] = []

            if cls.__name__ not in self.__cfls:
                self.__cfls[cls.__name__] = []

            for c in cls.__depends__:
                if c not in self.__deps:
                    self.__deps[c] = []
                self.__deps[c].append(cls.__name__)

            for c in cls.__conflicts__:
                if c not in self.__cfls:
                    self.__cfls[c] = []
                self.__cfls[c].append(cls.__name__)
                if c not in self.__cfls[cls.__name__]:
                    self.__cfls[cls.__name__].append(c)

        cl = self.config_list
        for name, cls in self.__classes.items():
            for dep in self.__deps[name]:
                # plugins that are required by not unloadable plugins are not unloadable too
                if not self.__classes[dep].__unloadable__:
                    cls.__unloadable__ = False

            if (cls.__autoload__ or (cl and cls.__name__ in cl)) and \
                    not (cls.__unloadable__ and cl and "!" + cls.__name__ in cl):
                self.__load_plugin(cls)

    def disable_plugin(self, plugin: str) -> bool:
        return False

    def enable_plugin(self, plugin: str) -> bool:
        return True

    def __load_plugin(self, cls: Type[_T]) -> None:
        if cls.__name__ in self.__loaded:
            return

        for dep in cls.__depends__:
            if dep not in self.__loaded:
                if dep not in self.__classes:
                    raise Exception(f"Could not satisfy dependency {cls.__name__} -> {dep}")
                try:
                    self.__load_plugin(self.__classes[dep])
                except Exception as e:
                    logging.exception(e)
                    raise

        for cfl in self.__cfls[cls.__name__]:
            if cfl in self.__classes:
                if self.__classes[cfl].__priority__ > cls.__priority__ and not self.disable_plugin(cfl) \
                        and not self.enable_plugin(cls.__name__):
                    logging.warning(f"Not loading {cls.__name__} because its conflict has higher priority")
                    return

            if cfl in self.__loaded:
                if cls.__priority__ > self.__classes[cfl].__priority__ and not self.enable_plugin(cfl):
                    self.unload_plugin(cfl)
                else:
                    raise LoadException(f"Not loading conflicting plugin {cls.__name__} due to lower priority")

        logging.info(f"loading {cls}")
        inst = cls(self.parent)
        try:
            inst._load()
        except Exception:
            logging.error(f"Failed to load {cls.__name__}", exc_info=True)
            if not cls.__unloadable__:
                bmexit()

            raise  # NOTE TO SELF: might cause bugs

        else:
            self.__plugins[cls.__name__] = inst

            self.__loaded.append(cls.__name__)
            self.emit("plugin-loaded", cls.__name__)

    def __getattr__(self, key: str) -> Any:
        try:
            return self.__plugins[key]
        except KeyError:
            return self.__dict__[key]

    def unload_plugin(self, name: str) -> None:
        if self.__classes[name].__unloadable__:
            for d in self.__deps[name]:
                self.unload_plugin(d)

            if name in self.__loaded:
                logging.info(f"Unloading {name}")
                try:
                    inst = self.__plugins[name]
                    inst._unload()
                except NotImplementedError:
                    logging.warning("Plugin cannot be unloaded")
                else:
                    self.__loaded.remove(name)
                    del self.__plugins[name]
                    self.emit("plugin-unloaded", name)

        else:
            raise Exception(f"Plugin {name} is not unloadable")

    def get_plugins(self) -> Dict[str, _T]:
        return self.__plugins

    _U = TypeVar("_U")

    def get_loaded_plugins(self, protocol: Type[_U]) -> Iterable[_U]:
        for name in self.__loaded:
            plugin = self.__plugins[name]
            if isinstance(plugin, protocol):
                yield plugin


class PersistentPluginManager(PluginManager):
    def __init__(self, plugin_class: Type[_T], module_path: ModuleType, parent: object) -> None:
        super().__init__(plugin_class, module_path, parent)

        self.__config = Config("org.blueman.general")

        self.__config.connect("changed::plugin-list", self.on_property_changed)

    def disable_plugin(self, plugin: str) -> bool:
        plugins = self.__config["plugin-list"]
        return "!" + plugin in plugins

    def enable_plugin(self, plugin: str) -> bool:
        plugins = self.__config["plugin-list"]
        return plugin in plugins

    def set_config(self, plugin: str, state: bool) -> None:
        plugins = self.__config["plugin-list"]
        if plugin in plugins:
            plugins.remove(plugin)
        elif "!" + plugin in plugins:
            plugins.remove("!" + plugin)

        plugins.append(str("!" + plugin) if not state else str(plugin))
        self.__config["plugin-list"] = plugins

    @property
    def config_list(self) -> List[str]:
        list: List[str] = self.__config["plugin-list"]
        return list

    def on_property_changed(self, config: Gio.Settings, key: str) -> None:
        for item in config[key]:
            disable = item.startswith("!")
            if disable:
                item = item.lstrip("!")

            try:
                cls: Type[BasePlugin] = self.get_classes()[item]
                if not cls.__unloadable__ and disable:
                    logging.warning(f"warning: {item} is not unloadable")
                elif item in self.get_loaded() and disable:
                    self.unload_plugin(item)
                elif item not in self.get_loaded() and not disable:
                    try:
                        self.load_plugin(item, user_action=True)
                    except Exception as e:
                        logging.exception(e)
                        self.set_config(item, False)

            except KeyError:
                logging.warning(f"warning: Plugin {item} not found")
                continue



import logging
from typing import Dict

from blueman.bluez.errors import BluezDBusException
from blueman.bluez.obex.Base import Base
from gi.repository import GObject, GLib

from blueman.bluemantyping import GSignals


class ObjectPush(Base):
    __gsignals__: GSignals = {
        'transfer-started': (GObject.SignalFlags.NO_HOOKS, None, (str, str,)),
        'transfer-failed': (GObject.SignalFlags.NO_HOOKS, None, (str,)),
    }

    _interface_name = 'org.bluez.obex.ObjectPush1'

    def __init__(self, obj_path: str):
        super().__init__(obj_path=obj_path)

    def send_file(self, file_path: str) -> None:
        def on_transfer_started(transfer_path: str, props: Dict[str, str]) -> None:
            logging.info(" ".join((self.get_object_path(), file_path, transfer_path)))
            self.emit('transfer-started', transfer_path, props['Filename'])

        def on_transfer_error(error: BluezDBusException) -> None:
            logging.error(f"{file_path} {error}")
            self.emit('transfer-failed', error)

        param = GLib.Variant('(s)', (file_path,))
        self._call('SendFile', param, reply_handler=on_transfer_started, error_handler=on_transfer_error)

    def get_session_path(self) -> str:
        path: str = self.get_object_path()
        return path



from gettext import gettext as _
from typing import Any, Optional

from blueman.bluez.Adapter import Adapter
from blueman.bluez.errors import DBusNoSuchAdapterError
from blueman.plugins.AppletPlugin import AppletPlugin
from gi.repository import GLib
import logging


class DiscvManager(AppletPlugin):
    __depends__ = ["Menu"]
    __author__ = "Walmis"
    __icon__ = "edit-find"
    __description__ = _(
        "Provides a menu item for making the default adapter temporarily visible when it is set to hidden by default")

    __gsettings__ = {
        "schema": "org.blueman.plugins.discvmanager",
        "path": None
    }
    __options__ = {
        "time": {
            "type": int,
            "default": 60,
            "name": _("Discoverable timeout"),
            "desc": _("Amount of time in seconds discoverable mode will last"),
            "range": (60, 600)
        }
    }

    adapter: Optional[Adapter]

    def on_load(self) -> None:
        self.item = self.parent.Plugins.Menu.add(self, 20, text=_("_Make Discoverable"), icon_name="edit-find",
                                                 tooltip=_("Make the default adapter temporarily visible"),
                                                 callback=self.on_set_discoverable, visible=False)
        self.adapter = None
        self.time_left = -1

        self.timeout: Optional[int] = None

    def on_unload(self) -> None:
        self.parent.Plugins.Menu.unregister(self)
        del self.item

        if self.timeout:
            GLib.source_remove(self.timeout)

    def on_manager_state_changed(self, state: bool) -> None:
        if state:
            self.init_adapter()
            self.update_menuitems()
        else:
            self.adapter = None
            self.update_menuitems()

    def on_update(self) -> bool:
        self.time_left -= 1
        self.item.set_text(_("Discoverable… %ss") % self.time_left)
        self.item.set_sensitive(False)

        return True

    def on_set_discoverable(self) -> None:
        if self.adapter:
            self.adapter.set("Discoverable", True)
            self.adapter.set("DiscoverableTimeout", self.get_option("time"))

    def init_adapter(self) -> None:
        try:
            self.adapter = self.parent.Manager.get_adapter()
        except DBusNoSuchAdapterError:
            self.adapter = None

    def on_adapter_added(self, path: str) -> None:
        if self.adapter is None:
            self.init_adapter()
            self.update_menuitems()

    def on_adapter_removed(self, path: str) -> None:
        logging.info(path)
        if self.adapter is None:
            # FIXME we appear to call this more than once on adapter removal
            logging.warning("Warning: adapter is None")
        elif path == self.adapter.get_object_path():
            self.init_adapter()
            self.update_menuitems()

    def on_adapter_property_changed(self, path: str, key: str, value: Any) -> None:
        if self.adapter and path == self.adapter.get_object_path():
            logging.debug(f"prop {key} {value}")
            if key == "DiscoverableTimeout":
                if value == 0:  # always visible
                    if self.timeout is not None:
                        GLib.source_remove(self.timeout)
                    self.time_left = -1
                    self.timeout = None
                else:
                    if self.time_left > -1:
                        if self.timeout is not None:
                            GLib.source_remove(self.timeout)
                    self.time_left = value

                    self.timeout = GLib.timeout_add(1000, self.on_update)
                    return

            elif (key == "Discoverable" and not value) or (key == "Powered" and not value):
                logging.info("Stop")
                if self.timeout is not None:
                    GLib.source_remove(self.timeout)
                self.time_left = -1
                self.timeout = None

            self.update_menuitems()

    def update_menuitems(self) -> None:
        if self.adapter is None:
            logging.warning("warning: Adapter is None")
            self.item.set_visible(False)
        elif (not self.adapter["Discoverable"] or self.adapter["DiscoverableTimeout"] > 0) and self.adapter["Powered"]:
            self.item.set_visible(True)
            self.item.set_text(_("_Make Discoverable"))
            self.item.set_sensitive(True)
        else:
            self.item.set_visible(False)
